"""
Implements data-based model parameters (:class:`ParameterSet`)

A :class:`ParameterSet` (or 'parset') is an intermediate representation of
model parameters. The main role of the parset is to store the calibration
values that are used to scale model parameters. Therefore, every parameter
in the model appears in the parset, not just the parameters in the databook.

"""

import io
from collections import defaultdict

import atomica.excel
import numpy as np
import pandas as pd

import sciris as sc
from .utils import NamedItem, TimeSeries
from .system import logger
import itertools
import json
import hashlib
from .version import version, gitinfo

__all__ = ["Parameter", "ParameterSet"]


class Parameter(NamedItem):
    """
    Class to hold one set of parameter values disaggregated by populations.

    :param name: The name of the parameter (should match framework code name)
    :param ts: A dict where the key is population name and the value is a TimeSeries instance

    """

    def __init__(self, name: str, ts: dict):
        NamedItem.__init__(self, name)
        self.ts = ts  # : Population-specific data is stored in TimeSeries within this dict, keyed by population name
        self.y_factor = sc.odict.fromkeys(self.ts, 1.0)  # : Calibration scale factors for the parameter in each population
        self.meta_y_factor = 1.0  # : Calibration scale factor for all populations
        self.skip_function = sc.odict.fromkeys(self.ts, None)  # : This can be a range of years [start,stop] between which the parameter function will not be evaluated
        self._interpolation_method = "linear"  #: Fallback interpolation method. It is _strongly_ recommended not to change this, but to call ``Parameter.smooth()` instead

    @property
    def pops(self):
        """
        Get populations contained in the Parameter

        :return: Generator/list of available populations

        """

        return self.ts.keys()

    def has_values(self, pop_name: str) -> bool:
        """
        Check if any values are present

        Returns True if this Parameter has values specified for the given population
        If the Parameter has an assumption, then the time value will be nan but a
        y-value will be present. If the Parameter normally has a function, then the
        y-value will be None. If a function Parameter has a scenario overwrite applied
        then actual values will be present. Essentially, if this function returns True,
        then the `interpolate()` method will return usable values

        :param pop_name: The code name of a population
        :return: ``True`` if any values are present for specified population, otherwise ``False``

        """

        if pop_name not in self.ts:
            return False
        elif self.ts[pop_name].has_data:
            return True
        else:
            return False

    def interpolate(self, tvec, pop_name: str) -> np.array:
        """
        Return interpolated parameter values for a given population

        The Parameter internally stores the interpolation method. The default is linear.
        It is possible to set it to 'pchip' or 'previous' or some other method. However,
        this would also be applied to any parameter scenarios that have modified the
        parameter and require interpolation. Therefore, it is STRONGLY recommended not to
        modify the fallback interpolation method, but to instead call `Parameter.smooth()`
        in advance with the appropriate options, if the interpolation matters.

        :param tvec: A scalar, list, or array or time values
        :param pop_name: The population to interpolate data for
        :return: An array with the interpolated values

        """

        return self.ts[pop_name].interpolate(tvec, method=self._interpolation_method)

    def sample(self, constant: bool) -> None:
        """
        Perturb parameter based on uncertainties

        This function modifies the parameter in-place. It would normally be called via
        :meth:`ParameterSet.sample()` which is responsible for copying this instance first.

        :param constant: If True, time series will be perturbed by a single constant offset. If False,
                         a different perturbation will be applied to each time specific value independently.

        """

        for k, ts in self.ts.items():
            self.ts[k] = ts.sample(constant)

    def smooth(self, tvec, method="smoothinterp", pop_names=None, **kwargs):
        """
        Smooth the parameter's time values

        Normally, Parameter instances contain temporally-sparse values from the databook.
        These are then interpolated to get the input parameter values at all model time points.
        The default interpolation is linear. However, sometimes it may be desired to make specific
        assumptions about the parameter value at intermediate times. These could be added directly
        to the Parameter's underlying `TimeSeries`.

        This method applies a smoothing method to the Parameter, modifying the underlying TimeSeries
        in-place. The operation is

        - Interpolated or smoothed values are generated for the requested times
        - All existing time points between and including the minimum and maximum values of `tvec` are removed
        - The time values generated in this function are inserted

        For example, to apply pchip interpolation to generate intermediate values

        >>> Parameter.smooth(P.settings.tvec, method='pchip',**kwargs)

        As this goes through `TimeSeries.interpolate` the same rules apply for the conditions under which
        the interpolation function gets used - specifically, interpolation is used if there are at least two
        finite time values present, otherwise, the assumption or single value will be used.

        :param tvec: New time points to add to the TimeSeries
        :param method: Method for generation of smoothed/interpolated values, default uses sciris smoothinterp
        :param pop_names: Optionally specify a list of populations to modify
        :param kwargs: Optionally pass arguments to the generating function/class constructor
        :return:

        """

        pop_names = sc.promotetolist(pop_names)
        if not pop_names:
            pop_names = self.ts.keys()

        if sc.isstring(method):
            if method == "smoothinterp":
                # Generating function for smooth-interp interpolator
                def smoothinterp(x1, y1, **kwargs):
                    def fcn(x2):
                        return sc.smoothinterp(x2, x1, y1, **kwargs)

                    return fcn

                method = smoothinterp
            elif method in ["pchip", "linear", "previous"]:
                pass
            else:
                raise Exception("Unknown smoothing method")

        for pop in pop_names:
            ts = self.ts[pop]
            v2 = ts.interpolate(tvec, method=method, **kwargs)
            ts.remove_between((min(tvec), max(tvec)))
            for t, v in zip(tvec, v2):
                ts.insert(t, v)


class Initialization:
    """
    This class stores initial compartment sizes

    In some cases it may be desirable to explicitly set initial compartment sizes rather than
    having them calculated based on the databook values/characteristics. An example of this could
    be wanting to initialize the model using steady-state compartment sizes computed from a
    prior model run. This class facilitates storing/applying the initial compartment sizes as well
    as capturing metadata for validation purposes.
    """

    def __init__(self, values=None, year=None, init_y_factor_hash=None, dt=None):
        """
        Construct an Initialization with explicit initial values

        This function can be used to explicitly set initial compartment sizes. More typically, the initial compartment
        sizes would be drawn from a previous model run. In that case, construct the ``Initialization`` using the
        ``Initialization.from_result()`` method, passing in the parset and result. For users, this is typically handled
        via ``ParameterSet.set_initialization(result)`` which allows easily passing a result into the ``ParameterSet``
        to set the initialization in one step.

        :param values: Provide a dictionary of values with compartment sizes. Keys should be tuples with
                       (comp_name,pop_name) and values should be either scalars (for normal compartments) or
                       arrays (for timed compartments). The size of the arrays for timed compartments will reflect
                       both the duration of the timed compartment and the simulation step size (noting that if timed
                       compartments are being used, the ``Initialization`` instance will not be reusable if the
                       simulation step size is subsequently changed, and will need to be re-created using the new
                       step size).
        :param year: Optionally specify the year used to generate the initialization (for provenance)
        :param init_y_factor_hash: Optionally specify the y-factor hash. If supplied, this will be checked against the parset when the initialization is applied
        :param dt: Optionally specify the timestep used to generate the initialization (for provenance)
        """
        self.year = year
        self.init_y_factor_hash = init_y_factor_hash
        self.dt = dt
        self.values = values

    @classmethod
    def from_result(cls, res, parset=None, year=None):
        """
        Construct an initialization based on a Result

        This method is used to create an ``Initialization`` instance when the initial compartment sizes are
        drawn from the state of a previously-run model. This facilitates initializing the model after numerically
        converging to a steady state or after an initial transient has passed.

        :param res: An Atomica ``Result`` instance
        :param parset: Optionally specify a ``ParameterSet`` instance containing y-factor values. If provided, subsequent
                       use of the initialization will check if the y-factors have changed since the initialization was
                       saved and display a warning if so.
        :param year: Optionally specify the year to draw compartment sizes from in the result. If not provided, the last
                     time point will be used. The year must exactly match a year contained in the result.
        :return: A new ``Initialization`` instance
        """

        from atomica.model import TimedCompartment  # Avoid circular import

        if year is None:
            year = res.model.t[-1]
        elif year not in res.model.t:
            raise Exception(f"Year {year} was not present in the result")

        idx = np.nonzero(res.model.t == year)[0][0]
        values = dict()  #: Dictionary of values with either {'comp':val} or {'comp':[vals]} depending on

        for pop in res.model.pops:
            for comp in pop.comps:
                if isinstance(comp, TimedCompartment):
                    values[(comp.name, pop.name)] = comp._vals[:, idx]
                else:
                    values[(comp.name, pop.name)] = comp.vals[idx]

        self = cls(values)
        self.year = year  #: Record year from which the initialization was originally computed
        self.init_y_factor_hash = None if parset is None else self.hash_y_factors(res.model.framework, parset)  # Record a hash of the Y-factors used for initialization
        self.dt = res.dt
        return self

    @staticmethod
    def hash_y_factors(framework, parset) -> str:
        """
        Hash y-factors used for initialization

        This method calculates a hash of the y-factors for the purpose of identifying if they have
        changed since the Initialization was originally created.

        :param framework: An ``at.Framework`` instance containing setup weights for compartments/characteristics
        :param parset: An ``at.ParameterSet`` instance containing y-factors
        :return: A hash computed from the y-factors used for normal compartment initialization
        """

        init_quantities = {k for k, v in framework.comps["setup weight"].to_dict().items() if v > 0}
        init_quantities.update(k for k, v in framework.characs["setup weight"].to_dict().items() if v > 0)

        d = {}
        for quantity in init_quantities:
            d[quantity] = dict(parset.pars[quantity].y_factor)
            d[quantity]["_meta_y_factor"] = parset.pars[quantity].meta_y_factor  # This might fail if there actually was a population called '_meta_y_factor' but that seems unlikely...

        return hashlib.sha256(json.dumps(d, sort_keys=True).encode("utf-8")).hexdigest()

    def apply(self, pop, framework=None, parset=None) -> None:
        """
        Insert saved values into compartments

        The Initialization may contain a hash of the y-factors for initialization compartments/characteristics that were
        used when the Initialization was generated, if the Initialization was generated based on a previous Result.
        When applying the initialization, if a framework and parset are provided, then the y-factors will be checked to
        determine if there are any differences in the initialization y-factors or not, and a warning will be displayed
        if so.

        :param pop: An at.model.Population instance
        :param framework: Optionally specify a framework object containing the specification of which comps/characs are used for initialization
        :param parset: Optionally specify the requested parset for initialization containing Y-factors to compare to the saved Y-factors
        """
        from atomica.model import TimedCompartment  # Avoid circular import

        # Check the y-factors
        if self.init_y_factor_hash is not None and framework is not None and parset is not None:
            init_y_factor_hash = self.hash_y_factors(framework, parset)
            if init_y_factor_hash != self.init_y_factor_hash:
                logger.warning("Y-factors used for initialization have changed since the saved initialization was generated. These Y-factors will have no effect because a saved initialization is being used. Remove the initialization by setting `Parset.initialization=None` to return to using the Y-factors to adjust the initialization.")

        for comp in pop.comps:
            if isinstance(comp, TimedCompartment):
                if (comp.name, pop.name) not in self.values:
                    comp._vals[:, 0] = 0
                else:
                    vals = sc.promotetoarray(self.values[(comp.name, pop.name)])
                    if comp._vals.shape[0] != vals.shape[0]:
                        if np.all(self.values[(comp.name, pop.name)] == 0):
                            # If there is a mismatch between the saved initialization duration and the
                            # duration for the current simulation AND if the values were all zero it is probably
                            # safe to assume the values can remain zero - this considerably increases flexibility in usage
                            comp._vals[:, 0] = 0
                        else:
                            # Otherwise, although we could try and guess how to assign the values, there is a risk that the time points
                            # are different because the step size was changed rather than the duration being changed. Therefore,
                            # if the sizes don't match and any values are nonzero, simply raise an error
                            raise Exception(f'The saved initialization for "{comp.name}" ({pop.name}) has {len(self.values[(comp.name, pop.name)])} time points, but the current parameters lead to a timed compartment duration with {comp._vals.shape[0]} time points. As nonzero values are present, the initialization cannot be applied.')
                    else:
                        comp._vals[:, 0] = self.values[(comp.name, pop.name)]
            else:
                if (comp.name, pop.name) not in self.values:
                    comp.vals[0] = 0
                else:
                    comp.vals[0] = self.values[(comp.name, pop.name)]

    def to_excel(self, writer):
        max_len = max(len(v) if not np.isscalar(v) else 1 for v in self.values.values())
        d = {}
        for k, v in self.values.items():
            d[k] = np.full(max_len, fill_value=np.nan)
            if np.isscalar(v):
                d[k][0] = v
            else:
                d[k][: len(v)] = v

        values = pd.DataFrame(d).T

        metadata = {
            "year": self.year,
            "init_y_factor_hash": self.init_y_factor_hash,
            "dt": self.dt,
        }
        metadata = pd.DataFrame.from_dict(metadata, orient="index")

        metadata.to_excel(writer, sheet_name="Initialization", startcol=0, startrow=0, index=True, header=False)
        values.to_excel(writer, sheet_name="Initialization", startcol=0, startrow=len(metadata) + 1, index=True, header=False)

    def __repr__(self):
        return sc.prepr(self)

    @classmethod
    def from_excel(cls, excelfile: pd.ExcelFile):
        """
        Construct an initialization from saved spreadsheet

        :param excelfile: A ``pd.ExcelFile`` containing an 'Initialization' sheet
        :return:
        """
        # excelfile = spreadsheet.pandas()

        metadata, value_df = atomica.excel.read_dataframes(excelfile.book["Initialization"])

        values = {}
        for k, s in value_df.T.reset_index().T.set_index([0, 1]).iterrows():
            v = s.dropna().values
            values[k] = v[0] if len(v) == 1 else v

        self = cls(values)

        for k, v in metadata.T.reset_index().T.set_index(0)[1].to_dict().items():
            setattr(self, k, v)

        return self


class ParameterSet(NamedItem):
    """
    Collection of model parameters to run a simulation

    A ParameterSet contains a collection of Parameters required to run the simulation.
    The parameters contain scale factors used for calibration, so often a project will
    contain multiple ParameterSets corresponding to different calibrations.

    Although parameters are constructed from ProjectData, there are two key differences

        - The ParameterSet contains calibration scale factors
        - The ParameterSet expands transfers and interactions into per-population parameters
          so they are stored on an equal basis (whereas ProjectData segregates them in
          :class:`TimeDependentValuesEntry` and :class:`TimeDependentConnections` due to
          the difference in how they are formatted in the databook)

    :param framework: A :class:`ProjectFramework` instance
    :param data: A :class:`ProjectData` instance
    :param name: Optionally specify the name of the parset

    """

    def __init__(self, framework, data, name="default"):
        NamedItem.__init__(self, name)
        self.version = version  # Track versioning information
        self.gitinfo = gitinfo

        self.pop_names = data.pops.keys()  # : List of all population code names contained in the ``ParameterSet``
        self.pop_labels = [pop["label"] for pop in data.pops.values()]  # : List of corresponding full names for populations
        self.pop_types = [pop["type"] for pop in data.pops.values()]  # : List of corresponding population types

        self.pars = sc.odict()  # : Stores the Parameter instances contained by this ParameterSet associated with framework comps, characs, and parameters
        self.transfers = sc.odict()  # : Stores the Parameter instances contained by this ParameterSet associated with databook transfers, keyed by source population
        self.interactions = sc.odict()  # : Stores the Parameter instances contained by this ParameterSet associated with framework interactions, keyed by source population

        self.initialization = None  #: Optionally store an ``Initialization`` instance to explicitly set initial compartment sizes

        # Instantiate all quantities that appear in the databook (compartments, characteristics, parameters)
        for name, tdve in data.tdve.items():
            ts = sc.odict()
            units = framework.get_databook_units(name).strip().lower()

            # First check units for all quantities
            for k, v in tdve.ts.items():
                if units != v.units.strip().lower():
                    message = f'The units for quantity "{framework.get_label(name)}" in the databook do not match the units in the framework. Expecting "{units}" but the databook contained "{ts.units.strip().lower()}"'
                    raise Exception(message)

            # Then populate the parameter time series
            for k in self.pop_names:
                if k in tdve.ts:
                    ts[k] = tdve.ts[k].copy()
                elif "all" in tdve.ts:
                    ts[k] = tdve.ts["all"].copy()
                elif "All" in tdve.ts:
                    ts[k] = tdve.ts["All"].copy()

            self.pars[name] = Parameter(name, ts)  # Keep only valid populations (discard any extra ones here)

        # Instantiate compartments and characteristics that have a default value and do not appear in the databook
        # Note that the default value will be used in all populations of the appropriate type
        # For the moment, framework validation ensures the default value is 0 - to avoid the confusion of default values being invisibly inserted
        # This requirement could be dropped in future if the use cases end up being unambiguous
        for _, spec in itertools.chain(framework.comps.iterrows(), framework.characs.iterrows()):
            if pd.isna(spec["databook page"]) and not pd.isna(spec["default value"]):
                assert spec.name not in self.pars, f"Quantity '{spec.name}' is not marked in the framework as having a databook page and it has a default value, but it has been loaded into the ParameterSet as data. If the quantity needs to be populated from the databook, either provide a databook page or remove the default value"
                ts = dict()
                units = framework.get_databook_units(name).strip().lower()
                for pop_name in self.pop_names:
                    if data.pops[pop_name]["type"] == spec["population type"]:
                        ts[pop_name] = TimeSeries(units=units, assumption=spec["default value"])
                self.pars[spec.name] = Parameter(spec.name, ts)

        # Instantiate parameters not in the databook
        for _, spec in framework.pars.iterrows():
            if spec.name not in self.pars:
                ts = dict()
                units = framework.get_databook_units(spec.name).strip().lower()
                for pop_name in self.pop_names:
                    if data.pops[pop_name]["type"] == spec["population type"]:
                        ts[pop_name] = TimeSeries(units=units)
                self.pars[spec.name] = Parameter(spec.name, ts)

        # Instantiate parameters for transfers and interactions
        # As with TDVE quantities, these must be initialized from the databook
        for tdc in data.transfers + data.interpops:
            if tdc.type == "transfer":
                item_storage = self.transfers
            elif tdc.type == "interaction":
                item_storage = self.interactions
            else:
                raise Exception("Unknown time-dependent connection type")

            name = tdc.code_name  # The name of this interaction e.g. 'age'
            item_storage[name] = sc.odict()
            ts_dict = defaultdict(dict)

            for pop_link, ts in tdc.ts.items():
                ts_dict[pop_link[0]][pop_link[1]] = ts.copy()

            for source_pop, ts in ts_dict.items():
                item_storage[name][source_pop] = Parameter(name + "_from_" + source_pop, sc.dcp(ts))

    def __setstate__(self, d):
        from .migration import migrate

        self.__dict__ = d
        parset = migrate(self)
        self.__dict__ = parset.__dict__

    def all_pars(self):
        """
        Return an iterator over all Parameters

        This is useful because transfers and interaction :class:`Parameter` instances are stored in
        nested dictionaries, so it's not trivial to iterate over all of them.

        :return: Generator over all :class:`Parameter` instances contained in the ``ParameterSet``

        """

        for par in self.pars.values():
            yield par
        for obj in self.transfers.values() + self.interactions.values():
            for par in obj.values():
                yield par

    def get_par(self, name: str, pop: str = None) -> Parameter:
        """
        Retrieve parameter instance

        The parameter values for interactions and transfers are stored keyed by
        the source/from population. Thus, if the quantity name is an interaction
        or transfer, it is also necessary to specify the source population in order
        to return a :class:`Parameter` instance. However, transfer parameters can
        also be identified by a parameter name "<transfer_name>_from_<pop>" such
        that ``ParameterSet.get_par('age','5-14')`` is equivalent to
        ``ParameterSet.get_par('age_from_5-14')``.

        :param name: The code name of a parameter, interaction, or transfer
        :param pop:
        :return: A :class:`Parameter` instance
        """

        if name in self.pars:
            assert pd.isna(pop), f'"{name}" is a parameter so the ``pop`` should not be specified'
            return self.pars[name]
        elif name in self.transfers:
            assert not pd.isna(pop), f'"{name}" is a transfer, so the ``pop`` must be specified'
            return self.transfers[name][pop]
        elif name in self.interactions:
            assert not pd.isna(pop), f'"{name}" is an interaction, so the ``pop`` must be specified'
            return self.interactions[name][pop]
        else:
            for transfer in self.transfers.values():
                for par in transfer.values():
                    if par.name == name:
                        return par

        raise KeyError(f'Parameter "{name}" not found')

    def sample(self, constant=True):
        """
        Return a sampled copy of the ParameterSet

        :param constant: If True, time series will be perturbed by a single constant offset. If False,
                         a different perturbation will be applied to each time specific value independently.
        :return: A new :class:`ParameterSet` with perturbed values

        """

        new = sc.dcp(self)
        for par in new.all_pars():
            par.sample(constant)
        return new

    def make_constant(self, year: float):
        """
        Return a constant copy of the ParameterSet

        This function will return a copy of the ParameterSet where all parameter time series have been replaced
        with temporally static versions.

        :param year: Year to use for interpolation
        :return: A copy of the ParameterSet with constant parameters
        """
        ps = self.copy(f"{self.name} (constant)")
        for par in ps.all_pars():
            for ts in par.ts.values():
                ts.insert(None, ts.interpolate(year))
                ts.t = []
                ts.vals = []
        return ps

    # SAVE AND LOAD CALIBRATION (Y-FACTORS)

    @property
    def y_factors(self) -> dict:
        """
        Return y-values in a dictionary

        Note that any missing populations reflect population types. For example, the
        dictionary for a parameter in one population type will not contain any entries
        for populations that belong to another type

        :return: Dictionary keyed by ``(par, pop)`` containing a dict of y-values
        """
        y_factors = {}

        for par_name, par in self.pars.items():
            y_factors[(par_name, None)] = sc.mergedicts({"meta_y_factor": par.meta_y_factor}, par.y_factor)

        for par_name, d in self.interactions.items() + self.transfers.items():
            for pop_name, par in d.items():
                y_factors[(par_name, pop_name)] = sc.mergedicts({"meta_y_factor": par.meta_y_factor}, par.y_factor)

        return y_factors

    def calibration_spreadsheet(self) -> sc.Spreadsheet:
        """
        Return y-values in a spreadsheet

        Note that the tabular structure contains missing entries for any interactions
        that don't exist (e.g. due to missing population types) - these can be identified
        with ``pd.isna``

        :return: Spreadsheet containing y-factors in tabular form
        """

        # Initialize the bytestream
        f = io.BytesIO()

        writer = pd.ExcelWriter(f, engine="xlsxwriter")
        # writer.book.set_properties({"category": "atomica:framework"})
        # standard_formats(writer.book)  # Apply formatting

        # worksheet = writer.book.add_worksheet("Y-factors")

        df = pd.DataFrame(self.y_factors).T
        df.index.set_names(["par", "pop"], inplace=True)
        df.to_excel(writer, sheet_name="Y-factors", merge_cells=False)  # Write index if present

        if self.initialization:
            self.initialization.to_excel(writer)

        # Close the workbook
        writer.close()

        return sc.Spreadsheet(f)

    def set_initialization(self, res, year=None):
        self.initialization = Initialization.from_result(res, parset=self, year=year)

    def apply_initialization(self, pop, framework=None) -> None:
        """

        :param pop: An ``at.Model.Population`` instance containing compartment objects that require initialization
        :param framework: A Framework containing a specification of which compartments/characteristics are ordinarily
                          used for initialization. If a framework is not provided, then the y-factors will not be
                          validated.
        :return:
        """
        if self.initialization is None:
            raise Exception("Attempted to apply an explicit compartment initialization but the parset does not contain an initialization")

        self.initialization.apply(pop, framework, self)

    def save_calibration(self, fname) -> None:
        """
        Save y-values to file

        :param fname: ``str`` or ``Path`` specifying location of file to save

        """
        ss = self.calibration_spreadsheet()
        ss.save(fname)

    def load_calibration(self, spreadsheet: sc.Spreadsheet) -> None:
        """
        Load calibration y-factors

        This function reads a spreadsheet created by ``ParameterSet.save_calibration()`` and
        inserts the y-factors. It is permissive in that

        - If y-factors are present in the spreadsheet and not in the ``ParameterSet`` then they
          will be skipped
        - If y-factors are missing in the spreadsheet, the existing values will be maintained

        :param spreadsheet:
        :return:
        """
        if not isinstance(spreadsheet, sc.Spreadsheet):
            spreadsheet = sc.Spreadsheet(spreadsheet)

        excelfile = spreadsheet.pandas()

        df = pd.read_excel(excelfile, "Y-factors" if "Y-factors" in excelfile.sheet_names else 0)
        df.set_index(["par", "pop"], inplace=True)

        if df.index.duplicated().any():
            msg = "The calibration file contained duplicate entries:"
            for par, pop in df.index[df.index.duplicated()].unique():
                msg += f'\n\t- {par} ({"meta/all populations" if pd.isna(pop) else pop})'
            raise Exception(msg)

        for (par_name, pop_name), values in df.to_dict(orient="index").items():
            try:
                par = self.get_par(par_name, pop_name)
            except KeyError:
                if pop_name:
                    logger.debug(f"{par.name} in {pop_name} was not found, ignoring y-factors for this quantity")
                else:
                    logger.debug(f"{par.name} was not found, ignoring y-factors for this quantity")
                continue

            for k, v in values.items():
                if pd.isna(v):
                    continue

                if k == "meta_y_factor":
                    par.meta_y_factor = v
                else:
                    if k in par.y_factor:
                        par.y_factor[k] = v
                    else:
                        logger.debug(f"The ParameterSet does not define a y-factor for {par.name} in the {k} population (e.g., because the population type does not match the parameter) - skipping")

        if "Initialization" in excelfile.sheet_names:
            self.initialization = Initialization.from_excel(excelfile)
        else:
            self.initialization = None

        logger.debug("Loaded calibration from file")



"""
Check whether automated model documentation template generation works
"""

import atomica as at
import numpy as np


def test_program_scenarios():

    P = at.demo("tb", do_run=False)

    # Get the default values for coverage etc.
    instructions = at.ProgramInstructions(2018)
    res_baseline = P.run_sim(result_name="Baseline", parset="default", progset="default", progset_instructions=instructions)

    alloc = res_baseline.get_alloc(2018)
    capacity = res_baseline.get_coverage("capacity", 2018)
    coverage = res_baseline.get_coverage("fraction", 2018)

    # Run a budget scenario manually
    doubled_budget = {x: v * 2 for x, v in alloc.items()}
    instructions = at.ProgramInstructions(2018, alloc=doubled_budget)
    res_doubled = P.run_sim(result_name="Doubled budget", parset="default", progset="default", progset_instructions=instructions)

    # Compare spending in 2018
    d = at.PlotData.programs([res_baseline, res_doubled], quantity="spending")
    d.interpolate(2018)
    at.plot_bars(d, stack_outputs="all")

    # Run a capacity scenario manually
    doubled_capacity = {x: v * 2 for x, v in capacity.items()}
    instructions = at.ProgramInstructions(2018, capacity=doubled_capacity)
    res_capacity = P.run_sim(result_name="Doubled capacity", parset="default", progset="default", progset_instructions=instructions)

    # # Compare capacity in 2018
    d = at.PlotData.programs([res_baseline, res_capacity], quantity="coverage_capacity")
    d.interpolate(2018)
    at.plot_bars(d, stack_outputs="all")

    # Run a coverage scenario manually
    doubled_coverage = {x: v * 2 for x, v in coverage.items()}
    instructions = at.ProgramInstructions(2018, coverage=doubled_coverage)
    res_coverage = P.run_sim(result_name="Doubled coverage", parset="default", progset="default", progset_instructions=instructions)

    # Compare coverage in 2018 - notice how the output coverage is capped to 1.0
    # even though the instructions contain fractional coverage values >1.0
    d = at.PlotData.programs([res_baseline, res_coverage], quantity="coverage_fraction")
    d.interpolate(2018)
    at.plot_bars(d, stack_outputs="all")

    # Compare program outcomes (incidence from 2018-2023)
    # Note that the doubled capacity scenario here is basically the same as
    # the doubled budget scenario because there were no capacity constraints
    # (the main use case for running a capacity scenario would be to
    # investigate circumventing capacity constraints).
    # On the other hand, the coverage scenario has fixed coverage from 2018-2023 whereas
    # the other scenarios have variable coverage, which is why the coverage scenario has
    # a different outcome
    d = at.PlotData([res_baseline, res_doubled, res_capacity, res_coverage], outputs=":acj", pops="total", t_bins=[2018, 2023])
    # Show change in incidence relative to baseline to improve clarity in this plot
    baseline = d.series[0].vals[0]
    for s in d.series:
        s.vals -= baseline
    at.plot_bars(d)

    # Run a budget scenario via the actual scenario infrastructure
    scen = at.BudgetScenario(name="Doubled budget scenario", alloc=doubled_budget, start_year=2018)
    res_doubled_scen = scen.run(P, parset="default", progset="default")

    # Run a coverage scenario via the scenario infrastructure
    scen = at.CoverageScenario(name="Double coverage scenario", coverage=doubled_coverage, start_year=2018)
    res_coverage_scen = scen.run(P, parset="default", progset="default")

    # Check that the infrastructure gives the same result as direct instructions and
    # also that the budget and coverage scenarios give different results
    d = at.PlotData([res_doubled, res_doubled_scen, res_coverage, res_coverage_scen], outputs=":acj", pops="total", t_bins=[2018, 2023])
    at.plot_bars(d)


def test_timevarying_progscen():
    # This test demonstrates doing time-varying overwrites
    # The example below shows how you can pass in a TimeSeries
    # instead of a scalar in the dict of overwrites. Although shown
    # for coverage below, this same approach works for alloc/budget
    # and capacity scenarios as well

    P = at.demo("sir", do_run=False)
    instructions = at.ProgramInstructions(2018)
    res_baseline = P.run_sim(result_name="Baseline", parset="default", progset="default", progset_instructions=instructions)

    coverage = {
        "Risk avoidance": 0.5,
        "Harm reduction 1": 0.25,
        "Harm reduction 2": at.TimeSeries([2018, 2020], [0.7, 0.2]),
    }
    scen = at.CombinedScenario("Reduced coverage", parsetname="default", progsetname="default", instructions=at.ProgramInstructions(coverage=coverage, start_year=2018))
    scen_result = scen.run(project=P)
    d = at.PlotData.programs([res_baseline, scen_result], quantity="coverage_fraction")
    at.plot_series(d)


def test_combined_scenario():

    P = at.demo("tb", do_run=False)

    res_baseline = P.run_sim(P.parsets[0], P.progsets[0], at.ProgramInstructions(2018))

    # Make a parameter overwrite]
    scenario_values = dict()
    scenario_values["b_rate"] = dict()
    scenario_values["b_rate"]["0-4"] = dict()
    scenario_values["b_rate"]["0-4"]["t"] = [2016.0, 2020.0, 2050]
    scenario_values["b_rate"]["0-4"]["y"] = [270000, 300000, 300000]

    # Make instructions for the program scenario
    alloc = {"BCG": at.TimeSeries([2018, 2020], [345000, 500000])}
    coverage = {"PCF": at.TimeSeries([2018, 2020], [0.00274411, 0.004])}
    instructions = at.ProgramInstructions(2018, alloc=alloc, coverage=coverage)

    # Instantiate the combined scenario
    scen = at.CombinedScenario(name="Combined test", parsetname="default", progsetname="default", scenario_values=scenario_values, instructions=instructions)

    # Run the scenario via `Project.run_scenarios()` and check the output
    for s in P.scens.values():
        s.active = False
    P.scens.append(scen)
    scen_result = P.run_scenarios(store_results=False)[0]

    # Check parameter overwrite
    d = at.PlotData([res_baseline, scen_result], outputs="b_rate", pops="0-4")
    at.plot_series(d, axis="results")

    # Check budget overwrite
    d = at.PlotData.programs([res_baseline, scen_result], quantity="spending", outputs="BCG")
    at.plot_series(d, axis="results")

    # Check coverage overwrite
    d = at.PlotData.programs([res_baseline, scen_result], quantity="coverage_fraction", outputs="PCF")
    at.plot_series(d, axis="results")

    # Check the combined scenario would work for just parameters or just programs
    pars_only = at.CombinedScenario(name="Parameters only", scenario_values=scenario_values)
    pars_only.run(project=P, parset=P.parsets["default"])

    progs_only = at.CombinedScenario(name="Programs only", instructions=instructions)
    progs_only.run(project=P, parset=P.parsets["default"], progset=P.progsets["default"])


def test_parameter_scenarios():

    proj = at.demo("sir", do_run=False)
    proj.settings.update_time_vector(start=2000, end=2023)

    # Check that it runs with an empty scvalues
    scvalues = dict()
    scen = proj.make_scenario(which="parameter", scenario_values=scvalues)
    scen.run(proj, proj.parsets["default"])

    # Check that it runs with a single overwrite
    scen_par1 = "contacts"
    scen_pop = "adults"
    scvalues[scen_par1] = dict()
    scvalues[scen_par1][scen_pop] = dict()
    scvalues[scen_par1][scen_pop]["y"] = [80.0, 40]
    scvalues[scen_par1][scen_pop]["t"] = [2010.0, 2020.0]
    scen = proj.make_scenario(which="parameter", scenario_values=scvalues)
    scen_results = scen.run(proj, proj.parsets["default"])

    # Check that default is linear interpolation
    var = scen_results.get_variable(scen_par1, scen_pop)[0]
    assert np.allclose(var.vals[var.t == 2010][0], 80, equal_nan=True)  # Default tolerances are rtol=1e-05, atol=1e-08
    assert np.allclose(var.vals[var.t == 2015][0], 60, equal_nan=True)
    assert np.allclose(var.vals[var.t == 2020][0], 40, equal_nan=True)

    # Check stepped/previous method also works
    scen.interpolation = "previous"
    scen_results = scen.run(proj, proj.parsets["default"])
    var = scen_results.get_variable(scen_par1, scen_pop)[0]
    assert np.allclose(var.vals[var.t == 2018][0], 80, equal_nan=True)
    assert np.allclose(var.vals[var.t == 2019][0], 80, equal_nan=True)
    assert np.allclose(var.vals[var.t == 2020][0], 40, equal_nan=True)

    # Check that multiple overwrites work
    scen_par2 = "transpercontact"
    scen_pop = "adults"
    scvalues[scen_par2] = dict()
    scvalues[scen_par2][scen_pop] = dict()
    scvalues[scen_par2][scen_pop]["y"] = [0.008, 0.006]
    scvalues[scen_par2][scen_pop]["t"] = [2010.0, 2020.0]
    scen = proj.make_scenario(which="parameter", scenario_values=scvalues)
    scen_results = scen.run(proj, proj.parsets["default"])
    var1 = scen_results.get_variable(scen_par1, scen_pop)[0]
    var2 = scen_results.get_variable(scen_par2, scen_pop)[0]

    assert np.allclose(var1.vals[var1.t == 2010][0], 80, equal_nan=True)
    assert np.allclose(var1.vals[var1.t == 2015][0], 60, equal_nan=True)
    assert np.allclose(var1.vals[var1.t == 2020][0], 40, equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2010][0], 0.008, equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2015][0], 0.007, equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2020][0], 0.006, equal_nan=True)

    # Check that scenarios don't end
    scen_par1 = "contacts"
    scen_pop = "adults"
    scvalues[scen_par1] = dict()
    scvalues[scen_par1][scen_pop] = dict()
    scvalues[scen_par1][scen_pop]["y"] = [80.0, 40]
    scvalues[scen_par1][scen_pop]["t"] = [2010.0, 2020.0]
    scen = proj.make_scenario(which="parameter", scenario_values=scvalues)
    scen_results = scen.run(proj, proj.parsets["default"])
    var = scen_results.get_variable(scen_par1, scen_pop)[0]
    assert np.allclose(var.vals[var.t == 2020][0], 40, equal_nan=True)
    assert np.allclose(var.vals[var.t == 2021][0], 40, equal_nan=True)


def test_overwrite_function_scenario():
    proj = at.demo("sir", do_run=False)
    proj.settings.update_time_vector(start=2000, end=2023)
    baseline = proj.run_sim()

    baseline.get_variable("foi")[0].plot()

    # Check that it runs with a single overwrite
    scen_par1 = "foi"
    scen_pop = "adults"
    scvalues = dict()
    scvalues[scen_par1] = dict()
    scvalues[scen_par1][scen_pop] = dict()
    scvalues[scen_par1][scen_pop]["y"] = [0.1, 0.15]
    scvalues[scen_par1][scen_pop]["t"] = [2015.0, 2018.0]
    scen = proj.make_scenario(which="parameter", scenario_values=scvalues)
    scen_results = scen.run(proj, proj.parsets["default"])
    scen_results.get_variable("foi")[0].plot()

    var1 = baseline.get_variable(scen_par1, scen_pop)[0]
    var2 = scen_results.get_variable(scen_par1, scen_pop)[0]

    assert np.allclose(var1.vals[var1.t == 2010][0], var2.vals[var2.t == 2010][0], equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2015][0], 0.1, equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2018][0], 0.15, equal_nan=True)
    assert np.allclose(var2.vals[var2.t == 2019][0], 0.15, equal_nan=True)  # Check that the function didn't turn back on


def test_interaction_scenario():

    proj = at.demo("tb", do_run=False)
    parset = proj.parsets[0]
    res_baseline = proj.run_sim(result_name="Baseline", parset="default")

    # Run an interaction and transfer scenario - take out aging 0-4 to 5-14, and also
    # no FOI input for 15-64
    ps = at.ParameterScenario(name="examplePS")
    ps.add("age", ("0-4", "5-14"), [2020.0, np.inf], [0, 0])

    for from_pop in proj.data.pops:
        if from_pop in parset.interactions["w_ctc"] and "15-64" in parset.interactions["w_ctc"][from_pop].ts:
            ps.add("w_ctc", (from_pop, "15-64"), [2020.0, np.inf], [0, 0])

    res_scen = proj.run_sim(result_name="Scen", parset=ps.get_parset(parset, proj))

    d = at.PlotData([res_baseline, res_scen], "age_0-4_to_5-14", pops="0-4")
    at.plot_series(d, axis="results")

    d = at.PlotData([res_baseline, res_scen], "foi_in", pops="15-64")
    at.plot_series(d, axis="results")

    assert res_baseline.get_variable("foi_in", "15-64")[0].vals[-1] > 0
    assert res_scen.get_variable("foi_in", "15-64")[0].vals[-1] == 0  # Check scenario has 0 FOI in 15-64
    assert res_scen.get_variable("foi_in", "0-4")[0].vals[-1] > 0  # Check that a different population was not affected
    assert res_scen.get_variable("foi_in", "5-14")[0].vals[-1] > 0  # Check that a different population was not affected
    assert res_baseline.get_variable("age_0-4_to_5-14", "0-4")[0].vals[-1] > 0
    assert res_scen.get_variable("age_0-4_to_5-14", "0-4")[0].vals[-1] == 0  # Check scenario has 0 FOI in 15-64


if __name__ == "__main__":
    test_program_scenarios()
    test_timevarying_progscen()
    test_parameter_scenarios()
    test_combined_scenario()
    test_overwrite_function_scenario()
    test_interaction_scenario()



"""
Implements the automatic reconciliation algorithm

In any given model run, the parameters calculated by the :class:`ProgramSet`
should ideally match up with the values in the :class:`ParameterSet` so that
there are no discontinuities in parameter value. This may not be the case depending
on the data gathered and the calibration. Reconciliation aims to adjust the
internal parameters of the :class:`ProgramSet` to best match a :class:`ParameterSet`
in a particular year.

"""

import numpy as np
import sciris as sc
from .system import logger
from .system import FrameworkSettings as FS
import pandas as pd
import logging

__all__ = ["reconcile"]


def _extract_targets(result, progset, ti, eval_pars=None):
    # Store the target parset values in the same form as the progset computed parameter values
    # This will also mean that target_vals contains the par UIDs only for the impact parameters
    # requested for reconcilation (if None, then all will be used). For simplicity, just do it
    # for all programs and parameters, only the target ones will actually get looked at.
    #
    # INPUTS
    # - result : A result object for the baseline sim (it won't have a progset because it was run without programs)
    # - progset : The progset to use to get the target pars/pops and coverage denominator
    # - ti : Time indices to extract results from (could be several)
    # - eval_pars: Optional list of (par,pop) names for evaluation - otherwise, uses all impact pars

    # Get the parameter outcomes - assumes simulation end year was the reconciliation year (hence -1 index)
    target_vals = dict()
    for covout in progset.covouts.values():
        par_name, pop_name = covout.par, covout.pop
        if (eval_pars is None and result.framework.get_par(par_name)["targetable"] == "y") or (par_name, pop_name) in eval_pars:
            par = result.get_variable(par_name, pop_name)[0]
            if par.units == FS.QUANTITY_TYPE_NUMBER:
                # If a transition parameter in number units is being targeted, then the program outcome is in units of per person reached
                # at each timestep, while the parameter units are people/year. Thus, we need to convert the model parameter into the program
                # output units prior to reconciling
                target_vals[(par_name, pop_name)] = par.vals[ti] * result.dt / np.array([par.source_popsize(x) for x in ti])
            else:
                target_vals[(par_name, pop_name)] = par.vals[ti]

    # Get the coverage denominator in the reconciliation year (it's always the same, so can do it once here)
    num_eligible = {x: y[ti] for x, y in result.get_coverage("eligible").items()}

    return target_vals, num_eligible


def _update_progset(asd_vals, mapping, progset):
    # Updates progset in place, inserting values
    # INPUTS:
    # - asd_vals - asd array
    # - mapping - list of tuples specifying where to insert elements from the ASD array
    #             e.g.  [('unit_cost','BCG'),('outcome','v_rate','0-4','BCG')]
    #             The mapping is
    #             unit_cost - program
    #             capacity_constraint - program
    #             baseline - par,pop
    #             outcome - par,pop,program
    # - progset : ProgramSet to modify, should have only one time
    for x, target in zip(asd_vals, mapping):
        if target[0] == "unit_cost":
            assert len(progset.programs[target[1]].unit_cost.vals) == 1
            progset.programs[target[1]].unit_cost.vals[0] = x
        elif target[0] == "capacity_constraint":
            assert len(progset.programs[target[1]].capacity_constraint.vals) == 1
            progset.programs[target[1]].capacity_constraint.vals[0] = x
        elif target[0] == "baseline":
            progset.covouts[(target[1], target[2])].baseline = x
        elif target[0] == "outcome":
            progset.covouts[(target[1], target[2])].progs[target[3]] = x


def _prepare_bounds(progset, unit_cost_bounds, baseline_bounds, capacity_bounds, outcome_bounds):
    # This is a separate function to _prepare_asd_inputs() because there may be complex logic related to
    # defaults for constructing the bounds. At the end, the bounds dict contains upper and lower limits
    # for every parameter being considered
    #
    # OUTPUTS
    # bounds dict : {quantity:{key:(lower,upper)} where quantity is 'unit_cost','capacity_constraint','baseline','outcome', and key is:
    #     for unit_cost - program
    #     for capacity_constraint - program
    #     for baseline - par,pop
    #     for outcome - par,pop,program

    bounds = dict()
    bounds["unit_cost"] = dict()
    bounds["capacity_constraint"] = dict()
    bounds["baseline"] = dict()
    bounds["outcome"] = dict()

    for prog in progset.programs.values():
        if unit_cost_bounds:
            bounds["unit_cost"][prog.name] = prog.unit_cost.vals * np.array([1 - unit_cost_bounds, 1 + unit_cost_bounds])
        if capacity_bounds and prog.capacity_constraint.has_data:
            bounds["capacity_constraint"][prog.name] = prog.capacity_constraint.vals * np.array([1 - capacity_bounds, 1 + capacity_bounds])

    for covout in progset.covouts.values():
        if baseline_bounds:
            bounds["baseline"][(covout.par, covout.pop)] = covout.baseline * np.array([1 - baseline_bounds, 1 + baseline_bounds])
        if outcome_bounds:
            for prog, outcome in covout.progs.items():
                bounds["outcome"][(covout.par, covout.pop, prog)] = outcome * np.array([1 - outcome_bounds, 1 + outcome_bounds])

    return bounds


def _prepare_asd_inputs(progset, bounds):
    # Return initial values, upper-lower bounds, and mapping to insert the arrays into a progset
    # Only quantities that appear in the bounds dict will be reconciled
    x0 = list()
    xmin = list()
    xmax = list()
    mapping = list()

    for prog in progset.programs.values():
        if prog.name in bounds["unit_cost"]:  # Might need to check program_specific bounds here
            x0.append(prog.unit_cost.vals[0])
            xmin.append(bounds["unit_cost"][prog.name][0])
            xmax.append(bounds["unit_cost"][prog.name][1])
            mapping.append(("unit_cost", prog.name))
        if prog.name in bounds["capacity_constraint"]:  # Might need to check program_specific bounds here
            x0.append(prog.capacity_constraint.vals[0])
            xmin.append(bounds["capacity_constraint"][prog.name][0])
            xmax.append(bounds["capacity_constraint"][prog.name][1])
            mapping.append(("capacity_constraint", prog.name))

    for covout in progset.covouts.values():
        if (covout.par, covout.pop) in bounds["baseline"]:
            x0.append(covout.baseline)
            xmin.append(bounds["baseline"][(covout.par, covout.pop)][0])
            xmax.append(bounds["baseline"][(covout.par, covout.pop)][1])
            mapping.append(("baseline", covout.par, covout.pop))

            for prog, outcome in covout.progs.items():
                if (covout.par, covout.pop, prog) in bounds["outcome"]:
                    x0.append(outcome)
                    xmin.append(bounds["outcome"][(covout.par, covout.pop, prog)][0])
                    xmax.append(bounds["outcome"][(covout.par, covout.pop, prog)][1])
                    mapping.append(("outcome", covout.par, covout.pop, prog))

    return x0, xmin, xmax, mapping


def _objective(x, mapping, progset, eval_years, target_vals, num_eligible, dt):
    _update_progset(x, mapping, progset)  # Apply the changes to the progset
    capacities = progset.get_capacities(tvec=eval_years, dt=dt)  # Get number coverage using latest unit costs but default spending
    prop_coverage = progset.get_prop_coverage(tvec=eval_years, dt=dt, capacities=capacities, num_eligible=num_eligible)

    obj = 0.0
    for i in range(0, len(eval_years)):
        outcomes = progset.get_outcomes(prop_coverage=prop_coverage)
        for key in target_vals:  # Key is a (par,pop) tuple
            obj += (target_vals[key][i] - outcomes[key]) ** 2  # Add squared difference in parameter value
    return obj


def _convert_to_single_year(progset, reconciliation_year):
    # Take in a progset
    # Return a progset with values only in the reconciliation year
    # This is then what actually gets reconciled
    # And we can be sure that when we go to make changes in the reconciliation year, we do not
    # have any extra time points

    # Basically, we only need to modify program attributes
    reconciliation_year = sc.promotetoarray(reconciliation_year)
    new_progset = sc.dcp(progset)
    new_progset.tvec = reconciliation_year.copy()

    for prog in new_progset.programs.values():
        # NOTE - because of the TDVE format, we need to do the overwrite for ALL time series provided
        # (because they must all have the same time axis) even if they are not being modified by the reconciliation
        # This results in a clear mapping from the original progbook into the single-year format. Users can always
        # go back and modify the reconciled progbook
        if prog.spend_data.has_data:
            prog.spend_data.vals = prog.spend_data.interpolate(reconciliation_year)
            prog.spend_data.t = reconciliation_year.copy()
            prog.spend_data.assumption = None

        if prog.unit_cost.has_data:
            prog.unit_cost.vals = prog.unit_cost.interpolate(reconciliation_year)
            prog.unit_cost.t = reconciliation_year.copy()
            prog.unit_cost.assumption = None

        if prog.capacity_constraint.has_data:
            prog.capacity_constraint.vals = prog.capacity_constraint.interpolate(reconciliation_year)
            prog.capacity_constraint.t = reconciliation_year.copy()
            prog.capacity_constraint.assumption = None

        # This is tricky - maybe we do want to retain other values? Depends on what ends up happening with coverage
        if prog.coverage.has_data:
            prog.coverage.vals = prog.coverage.interpolate(reconciliation_year)
            prog.coverage.t = reconciliation_year.copy()
            prog.coverage.assumption = None

    return new_progset


# ASD takes in a list of values. So we need to map all of the things we are optimizing onto


def reconcile(project, parset, progset, reconciliation_year: float, max_time=10, unit_cost_bounds=0.0, baseline_bounds=0.0, capacity_bounds=0.0, outcome_bounds=0.0, eval_pars=None, eval_range=None):
    """
    Modify a progset to optimally match a parset in a specified year

    Reconciliation is a mapping from one progset to another. The output progset generates optimally matched output parameter
    values compared to the specified parset, in the reconciliation year for the progset's default spending. The output progset
    has internal attributes (such as unit costs) with values that are defined only in the reconciliation year. So while the
    input progset may have time varying unit costs etc. after reconciliation, they will be constant.

    The upper and lower bounds for unit costs, baseline, capacity_constraint, and outcome are specified as fractions of the initial value. For
    example, entering ``unit_cost_bounds=0.2`` would mean that unit costs would be allowed to range from 0.8 to 1.2 times the
    value in the input progset.

    :param project: A :class:`Project` instance
    :param parset: A :class:`ParameterSet` instance (or name of a parset contained in the project)
    :param progset: A :class:`ProgramSet` instance (or name of a progset contained in the project)
    :param reconciliation_year: Year to perform reconciliation in
    :param max_time: Optionally override the maximum execution time in ASD
    :param unit_cost_bounds: Optionally specify bounds for unit costs. Default is 0.0 (unit costs will not be changed)
    :param baseline_bounds: Optionally specify bounds for baseline spending. Default is 0.0 (no changes)
    :param capacity_bounds: Optionally specify bounds for capacity_constraint constraints. Default is 0.0 (no changes)
    :param outcome_bounds: Optionally specify bounds for outcome values. Default is 0.0 (no changes)
    :param eval_pars: Optionally select a subset of parameters for comparison. By default, all parameters overwritten by the progset will be used.
    :param eval_range: Optionally specify a range of years over which to evaluate the progset-parset match. By default, it will only use the reconciliation year
    :return: tuple containing
            - A reconciled :class:`ProgramSet` instance
            - A DataFrame comparing the unreconciled and reconciled progsets
            - A DataFrame comparing the parset parameters and progset

    """

    # Sanitize inputs
    parset = project.parset(parset)
    progset = project.progset(progset)

    logger.warning("Reconcilation when parameter is in number units not fully tested")

    reconciliation_year = sc.promotetoarray(reconciliation_year)
    assert len(reconciliation_year) == 1, "Reconciliation year must be a scalar"

    if eval_range is None:
        eval_range = [reconciliation_year[0], reconciliation_year[0] + project.settings.sim_dt]

    # Do a prerun to get the baseline values and coverage denominator
    parset_results = project.run_sim(parset=parset, progset=progset, store_results=False)
    ti = np.where((parset_results.model.t >= eval_range[0]) & (parset_results.model.t < eval_range[1]))[0]
    eval_years = parset_results.t[ti]
    target_vals, num_eligible = _extract_targets(parset_results, progset, ti, eval_pars)

    # Prepare ASD inputs
    new_progset = _convert_to_single_year(progset, reconciliation_year[0])
    bounds = _prepare_bounds(new_progset, unit_cost_bounds, baseline_bounds, capacity_bounds, outcome_bounds)  # Convert reconcile() inputs into full detailed bounds
    x0, xmin, xmax, mapping = _prepare_asd_inputs(new_progset, bounds)  # Assemble ASD variables

    logger.info("Reconciling in %.2f, evaluating from %.2f up to %.2f", reconciliation_year, eval_range[0], eval_range[1])

    args = {
        "mapping": mapping,
        "progset": new_progset,
        "eval_years": eval_years,
        "target_vals": target_vals,
        "num_eligible": num_eligible,
        "dt": project.settings.sim_dt,
    }

    optim_args = {
        # TODO - tune these variables specifically for reconciliation
        "xmin": xmin,
        "xmax": xmax,
        "verbose": 2,  # default verbosity
        "maxtime": max_time,
    }

    log_level = logger.getEffectiveLevel()
    if log_level < logging.WARNING:
        optim_args["verbose"] = 2
    else:
        optim_args["verbose"] = 0

    opt_result = sc.asd(_objective, x0, args, **optim_args)
    x_opt = opt_result["x"]

    _update_progset(x_opt, mapping, new_progset)  # Apply the changes to the progset

    # Before/after for quantities
    records = [(item[0], item[1:], orig_val, opt_val) for item, orig_val, opt_val in zip(mapping, x0, x_opt)]
    progset_comparison = pd.DataFrame.from_records(records, columns=["Quantity", "Identifier", "Before reconciliation", "After reconciliation"])

    # Before/after for parameters
    records = []
    old_capacities = progset.get_capacities(tvec=eval_years, dt=project.settings.sim_dt)
    new_capacities = new_progset.get_capacities(tvec=eval_years, dt=project.settings.sim_dt)
    old_prop_coverage = progset.get_prop_coverage(tvec=eval_years, dt=project.settings.sim_dt, capacities=old_capacities, num_eligible=num_eligible)
    new_prop_coverage = new_progset.get_prop_coverage(tvec=eval_years, dt=project.settings.sim_dt, capacities=new_capacities, num_eligible=num_eligible)
    for i, year in enumerate(eval_years):
        old_outcomes = progset.get_outcomes(prop_coverage={prog: cov[[i]] for prog, cov in old_prop_coverage.items()})  # Program outcomes for this year
        new_outcomes = new_progset.get_outcomes(prop_coverage={prog: cov[[i]] for prog, cov in new_prop_coverage.items()})  # Program outcomes for this year
        for (par, pop), target in target_vals.items():
            records.append((par, pop, year, target[0], old_outcomes[(par, pop)], new_outcomes[(par, pop)]))
    parameter_comparison = pd.DataFrame.from_records(records, columns=["Parameter", "Population", "Year", "Target", "Before reconciliation", "After reconciliation"])
    parameter_comparison["Difference"] = parameter_comparison["Before reconciliation"] - parameter_comparison["After reconciliation"]

    return new_progset, progset_comparison, parameter_comparison



# Test safe division in function_parser
import pytest
import atomica as at
import numpy as np
import matplotlib.pyplot as plt


@pytest.mark.parametrize("n_cols", [0, 1, 2])
@pytest.mark.parametrize("axis", ["outputs", "results", "pops"])
@pytest.mark.parametrize("legend_mode", ["together", "separate", ""])
def test_plot(n_cols, legend_mode, axis):

    P = at.demo("sir")
    d = at.PlotData(P.results[0], project=P)
    figs = at.plot_series(d, axis=axis, legend_mode=legend_mode, n_cols=n_cols)

    for fig in figs:
        plt.close(fig)



# CASCADE ENSEMBLES
# This script demonstrates how to use the CascadeEnsemble to plot cascades with uncertainty

import atomica as at
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
import sciris as sc


def test_ensemble_cascade():
    # BASIC SETUP
    testdir = at.parent_dir()

    P = at.Project(framework=testdir / "test_uncertainty_framework.xlsx", databook=testdir / "test_uncertainty_databook.xlsx")

    low_uncertainty_progset = at.ProgramSet.from_spreadsheet(testdir / "test_uncertainty_low_progbook.xlsx", project=P)
    high_uncertainty_progset = at.ProgramSet.from_spreadsheet(testdir / "test_uncertainty_high_progbook.xlsx", project=P)
    default_budget = at.ProgramInstructions(start_year=2018, alloc=low_uncertainty_progset)
    doubled_budget = default_budget.scale_alloc(2)

    # COMPARING YEARS
    #
    # Make a cascade plot comparing the same Results at different years, with
    # and without progsets
    ensemble = at.CascadeEnsemble(P.framework, "main")
    ensemble.run_sims(P, parset="default", n_samples=100)
    ensemble.plot_multi_cascade(years=[2018, 2020, 2023])
    plt.title("Default parset")

    ensemble.run_sims(P, parset="default", progset=low_uncertainty_progset, progset_instructions=default_budget, n_samples=100)
    ensemble.plot_multi_cascade(years=[2018, 2020, 2023])
    plt.title("Default progset (low uncertainty)")

    ensemble.run_sims(P, parset="default", progset=high_uncertainty_progset, progset_instructions=default_budget, n_samples=100)
    ensemble.plot_multi_cascade(years=[2018, 2020, 2023])
    plt.title("Default progset (high uncertainty)")

    # Demonstrate calling an inherited Ensemble plotting function - in this case, a basic
    # series plot showing cascade evolution over time. Note that the Ensemble was created
    # with only 3 time points, so the series plot also only contains those three times
    ensemble.plot_series(style="std", pops="m_rural")
    plt.legend()

    # COMPARING RESULTS
    #
    # We have a shortcut for comparing parameter scenarios - pass multiple instructions into `Ensemble.run_sims()` to automatically run them
    # A less common task would be to compare different progsets or parsets - these aren't built into `run_sims()` which is instead aimed at
    # the common task of testing different program scenarios against the same parset and progset samples, which is the most valid way
    # to examine uncertainty in differences (as illustrated in the final example in this file)

    ensemble = at.CascadeEnsemble(P.framework, "main", [2020, 2023])
    ensemble.run_sims(P, parset="default", progset=high_uncertainty_progset, progset_instructions=[default_budget, doubled_budget], n_samples=100, result_names=["Default", "Doubled"])

    ensemble.plot_multi_cascade(years=2020)
    plt.title("Spending comparison 2020")

    ensemble.plot_multi_cascade(years=2023)
    plt.title("Spending comparison 2023")

    # VIRTUAL STAGES
    #
    # In normal cascade plots, all cascade stages must consist of comps/characs only. Because `CascadeEnsemble` has a plotting
    # routine built around `PlotData` instead of `Result` objects, it is possible to introduce arbitrarily defined cascade stages.
    # This is aimed at advanced users of course, because the normal cascade plotting routine has validation checks like ensuring that
    # the cascade is properly nested, whereas these validation checks aren't performed here because of the arbitrary possibilities

    cascade = {"Prevalent": "all_people", "Screened": "all_screened", "Pre-diagnosed": "0.8*all_screened", "Diagnosed": "all_dx", "Treated": "all_tx", "Controlled": "all_con"}  # This 'stage' is not a compartment/characteristic at all

    ensemble = at.CascadeEnsemble(P.framework, cascade, [2020, 2023])
    ensemble.run_sims(P, parset="default", progset=high_uncertainty_progset, progset_instructions=[default_budget, doubled_budget], n_samples=100, result_names=["Default", "Doubled"])
    ensemble.plot_multi_cascade(years=2020)
    plt.title("Spending comparison 2020 with virtual stage")

    # CUSTOM MAPPING
    #
    # Make a cascade plot showing the difference between default and doubled spending
    # This demonstrates how to override the cascade ensemble's default mapping function
    # while still making use of `CascadeEnsemble.plot_multi_cascade`

    ensemble = at.CascadeEnsemble(P.framework, 0, [2020, 2023])  # Just put '0' to use the first cascade - doesn't matter which one since it gets overwritten below

    def cascade_difference(results):
        cascade_dict = at.sanitize_cascade(results[0].framework, "main")[1]  # Use `sanitize_cascade` to retrieve the cascade dictionary
        d1 = at.PlotData(results[0], outputs=cascade_dict)
        d2 = at.PlotData(results[1], outputs=cascade_dict)
        return d2 - d1

    ensemble.mapping_function = cascade_difference
    ensemble.run_sims(P, parset="default", progset=high_uncertainty_progset, progset_instructions=[default_budget, doubled_budget], n_samples=100, result_names=["Default", "Doubled"])
    ensemble.plot_multi_cascade(years=2023)
    plt.title("Difference between doubled budget and default budget")


if __name__ == "__main__":
    test_ensemble_cascade()



# Basic minimum working product test to verify all library models run
# Note that plotting is not tested here, because plotting (and other functions) already have
# separate automatic tests

import atomica as at
import os
import pytest
import numpy as np
import sciris as sc

testdir = at.parent_dir()
tmpdir = testdir / "temp"

# List available models based on which framework files exist
models = list()
for f in os.listdir(at.LIBRARY_PATH):
    if f.endswith("_framework.xlsx") and not f.startswith("~$"):
        models.append(f.replace("_framework.xlsx", ""))


def run_auto_calibration(proj):
    """Run an automatic calibration

    :param proj: A Project object
    :return:
    """

    proj.calibrate(max_time=10, new_name="auto")

    return


def run_parameter_scenario(proj):
    """Run an example parameter scenario for a project

    :param proj: A Project object
    :return: None

    """

    print("Testing parameter scenario")

    # Find a parameter from the databook
    for _, spec in proj.framework.pars.iterrows():
        if spec["databook page"]:
            break
    else:
        raise Exception("No Framework parameters appeared in the databook")

    par = proj.parsets[0].pars[spec.name]  # Get the parameter
    pop_name = par.pops[0]

    # Add overwrite with a slight increase in the 2020 value
    scvalues = dict()
    scvalues[spec.name] = dict()
    scvalues[spec.name][pop_name] = dict()
    scvalues[spec.name][pop_name]["t"] = [2015.0, 2020.0]
    scvalues[spec.name][pop_name]["y"] = par.interpolate(scvalues[spec.name][pop_name]["t"], pop_name) * np.array([1, 1.5]).ravel()
    scen = proj.make_scenario(which="parameter", name="Test", scenario_values=scvalues)
    res = scen.run(proj, proj.parsets["default"])
    if res.check_for_nans():
        raise Exception("NaNs detected")
    return


def run_reconciliation(proj):
    """Run an example reconciliation

    :param proj: A Project object
    :return: None

    """

    at.reconcile(project=proj, parset=proj.parsets[0], progset=proj.progsets[0], reconciliation_year=2018, unit_cost_bounds=0.05, baseline_bounds=0.05, capacity_bounds=0.05, outcome_bounds=0.05)

    return


def run_budget_scenario(proj):
    """Run a budget scenario

    :param proj: A Project object
    :return: None

    """

    print("Testing budget scenario")

    alloc = proj.progsets[0].get_alloc(2018)
    doubled_budget = {x: v * 2 for x, v in alloc.items()}
    instructions = at.ProgramInstructions(start_year=2018, alloc=doubled_budget)
    scen = at.CombinedScenario(name="Doubled budget", instructions=instructions)
    res = scen.run(proj, parset="default", progset="default")
    if res.check_for_nans():
        raise Exception("NaNs detected")

    return


def run_coverage_scenario(proj):
    """Run a coverage scenario

    :param proj: A Project object
    :return: None

    """

    print("Testing coverage scenario")

    half_coverage = {x: 0.5 for x in proj.progsets[0].programs.keys()}
    instructions = at.ProgramInstructions(start_year=2018, coverage=half_coverage)
    scen = at.CombinedScenario(name="Doubled budget", instructions=instructions)
    res = scen.run(proj, parset="default", progset="default")
    if res.check_for_nans():
        raise Exception("NaNs detected")

    return


def run_optimization(proj):
    """
    Run an optimization

    This demo optimization will adjust spending on all programs optimize all programs with the last default cascade stage)

    :param proj: A Project object
    :return: None

    """

    instructions = at.ProgramInstructions(alloc=proj.progsets[0], start_year=2020)  # Instructions for default spending
    adjustments = [at.SpendingAdjustment(x, 2020, "rel", 0.0, 2.0) for x in instructions.alloc.keys()]
    measurables = at.MaximizeCascadeStage(None, 2020)
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, maxtime=10)  # Evaluate from 2020 to end of simulation
    at.optimize(proj, optimization, parset=proj.parsets["default"], progset=proj.progsets["default"], instructions=instructions)

    return


def run_export(result, model):
    """
    Test exporting results

    :param result:
    :param model:
    :return:
    """

    at.export_results(result, tmpdir / f"{model}_single_export_test")  # Export single
    result.export_raw(tmpdir / f"{model}_raw_export_test")
    r2 = sc.dcp(result)
    r2.name = "Copied"
    at.export_results([result, r2], tmpdir / f"{model}_multiple_export_test")  # Export single


def run_regenerated_framework(proj):

    r1 = at.run_model(proj.settings, proj.framework, proj.parsets[0], proj.progsets[0], at.ProgramInstructions(start_year=2020))

    ss = proj.framework.to_spreadsheet()
    f2 = at.ProjectFramework(ss)

    r2 = at.run_model(proj.settings, f2, proj.parsets[0], proj.progsets[0], at.ProgramInstructions(start_year=2020))

    for p1, p2 in zip(r1.model.pops, r2.model.pops):
        for v1, v2 in zip(p1.pars + p1.comps + p1.characs + p1.links, p2.pars + p2.comps + p2.characs + p2.links):
            assert np.all(v1.vals == v2.vals)


# Testing optimizations and calibrations could be expensive
# Using the parametrize decorator means Pytest will treat
# each function call as a separate test, which can be run in parallel
# simply by adding `pytest -n 4` for instance. Or via tox
# `tox -- -o -n 4` (because the default config is to use n=2 for TravisCI)
@pytest.mark.parametrize("model", models)
def test_model(model):

    framework_file = at.LIBRARY_PATH / f"{model}_framework.xlsx"
    databook_file = at.LIBRARY_PATH / f"{model}_databook.xlsx"
    progbook_file = at.LIBRARY_PATH / f"{model}_progbook.xlsx"

    print("\n\nTESTING %s project" % (model))

    P = at.Project(framework=framework_file)

    if not os.path.isfile(databook_file):
        print("No databook found, test complete")
        return

    P.update_settings(sim_end=2025)  # Make sure we run until 2025

    # Test loading the databook and doing a basic run
    P.load_databook(databook_file, make_default_parset=True, do_run=True)

    # Test saving the databook (assume writing to disk works)
    P.data.to_spreadsheet()

    # Test exporting the calibration
    P.parsets[0].calibration_spreadsheet()

    # Test doing auto-calibration
    # run_auto_calibration(P)

    # Test a parameter scenario
    run_parameter_scenario(P)

    if not os.path.isfile(progbook_file):
        print("No progbook found, test complete")
        return

    # Test loading the progbook and doing a basic run
    P.load_progbook(progbook_file)
    res = P.run_sim(P.parsets[0], P.progsets[0], at.ProgramInstructions(start_year=2018))
    if res.check_for_nans():
        raise Exception("NaNs detected")

    # Test re-saving the framework
    run_regenerated_framework(P)

    # Test exporting the model with progset results
    run_export(res, model)

    # Test a reconciliation
    run_reconciliation(P)

    # Test a BudgetScenario
    run_budget_scenario(P)

    # Test a CoverageScenario
    run_coverage_scenario(P)

    # Test an optimization
    run_optimization(P)

    print("Test complete")


if __name__ == "__main__":

    np.seterr(all="raise")
    models = ["combined"]
    for m in models:
        test_model(m)



import numpy as np
import atomica as at
import os
import pytest

testdir = at.parent_dir()  # Must be relative to current file to work with tox


def test_no_compartments():
    # Test a framework with absolutely no compartments or transitions
    # Only parameters

    # Commands to generate the databook and progbook prior to being filled out

    f_path = testdir / "test_no_compartment_framework.xlsx"
    d_path = testdir / "test_no_compartment_databook.xlsx"
    p_path = testdir / "test_no_compartment_progbook.xlsx"
    # F = at.ProjectFramework(f_path)
    # D = at.ProjectData.new(F,[2000],pops=1,transfers=0)
    # D.save(d_path)
    # P = at.ProgramSet.new(tvec=[2000],framework=F,data=D,progs=1)
    # P.save(p_path)

    P = at.Project(name="test", framework=f_path)
    P.load_databook(d_path, do_run=False)
    P.load_progbook(p_path)
    P.settings.update_time_vector(dt=1)

    # Test a coverage scenario
    res_parset = P.run_sim("default")

    with pytest.raises(Exception):
        instructions = at.ProgramInstructions(2002)
        # Running a progset without compartments and without coverage in the instructions should result in an error
        P.run_sim(parset="default", progset="default", progset_instructions=instructions, result_name="Baseline (no progset)")

    instructions = at.ProgramInstructions(2002, coverage={"treatment": 0.6})
    res_progset = P.run_sim(parset="default", progset="default", progset_instructions=instructions, result_name="Increased treatment via progset")

    d = at.PlotData(results=[res_parset, res_progset], outputs=["deaths", {"Number treated": "p_treat*n_infections"}])
    at.plot_series(d, axis="results")

    # Check the values at the end of the simulation are as expected
    assert res_parset.get_variable("deaths")[0].vals[-1] == 40 * 0.1 + 60 * 0.4
    assert res_progset.get_variable("deaths")[0].vals[-1] == 60 * 0.1 + 40 * 0.4

    print("Test successfully completed")


if __name__ == "__main__":
    test_no_compartments()



"""
Implements various Optimizations in Atomica

This module implements the :class:`Optimization` class, which contains the
information required to perform an optimization in Atomica. An Optimization
effectively serves as a mapping from one set of program instructions to another.

"""

import logging
import pickle
from collections import defaultdict

import atomica
import numpy as np
import scipy.optimize

import sciris as sc
from .cascade import get_cascade_vals
from .model import Model, Link
from .parameters import ParameterSet
from .programs import ProgramSet, ProgramInstructions
from .results import Result
from .system import logger, NotFoundError
from .utils import NamedItem
from .utils import TimeSeries

__all__ = ["InvalidInitialConditions", "UnresolvableConstraint", "FailedConstraint", "Adjustable", "Adjustment", "SpendingAdjustment", "StartTimeAdjustment", "ExponentialSpendingAdjustment", "SpendingPackageAdjustment", "PairedLinearSpendingAdjustment", "Measurable", "MinimizeMeasurable", "MaximizeMeasurable", "AtMostMeasurable", "AtLeastMeasurable", "IncreaseByMeasurable", "DecreaseByMeasurable", "MaximizeCascadeStage", "MaximizeCascadeConversionRate", "Constraint", "TotalSpendConstraint", "Optimization", "optimize"]


class InvalidInitialConditions(Exception):
    """
    Invalid initial parameter values

    This error gets thrown if the initial conditions yield an objective value
    that is not finite

    """

    pass


class UnresolvableConstraint(Exception):
    """
    Unresolvable (ill-posed) constraint

    This error gets thrown if it is _impossible_ to satisfy the constraints. There are
    two modes of constraint failure
    - The constraint might not be satisfied on this iteration, but could be satisfied by other
      parameter values
    - The constraint is impossible to satisfy because it is inconsistent (for example, if the
      total spend is greater than the sum of the upper bounds on all the individual programs)
      in which case the algorithm cannot continue
    This error gets raised in the latter case, while the former should result in the iteration
    being skipped

    """

    pass


class FailedConstraint(Exception):
    """
    Not possible to apply constraint

    This error gets raised if a ``Constraint`` is unable to transform the instructions
    given the supplied parameter values (but other values may be acceptable). It signals
    that the algorithm should proceed immediately to the next iteration.

    """

    pass


class Adjustable:
    """
    Class to store single optimizable parameter

    An ``Adjustable`` represents a single entry in the ASD matrix. An ``Adjustment`` uses one or more
    ``Adjustables`` to make changes to the program instructions.

    :param name: The name of the adjustable
    :param limit_type: If bounds are provided, are they relative or absolute (``'abs'`` or ``'rel'``)
    :param lower_bound: Optionally specify minimum value
    :param upper_bound: Optionally specify maximum value
    :param initial_value: Optionally specify initial value. Most commonly, the initial value would be set by the ``Adjustment`` containing the ``Adjustable``

    """

    def __init__(self, name, limit_type="abs", lower_bound=-np.inf, upper_bound=np.inf, initial_value=None):
        self.name = name
        self.limit_type = limit_type
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.initial_value = initial_value

    def get_hard_bounds(self, x0: float = None) -> tuple:
        """
        Return hard bounds for the adjustable

        The hard bounds could be relative or absolute. If they are relative, then they are treated
        as being relative to the initial value for this adjustable. This is because not all adjustables
        can be directly drawn from the program instructions (e.g. parametric overwrites) in which case
        it would not be possible to extract initial values from the instructions alone. For consistency,
        all adjustables behave the same way with constraints relative to their initialization.

        :param x0: The reference value for relative constraints - only required if the limit type is not absolute
        :return: A tuple with ``(min,max)`` limits

        """

        self.lower_bound = self.lower_bound if self.lower_bound is not None else -np.inf
        self.upper_bound = self.upper_bound if self.upper_bound is not None else np.inf
        xmin = self.lower_bound if self.limit_type == "abs" else x0 * self.lower_bound
        xmax = self.upper_bound if self.limit_type == "abs" else x0 * self.upper_bound
        return xmin, xmax


class Adjustment:
    """
    Class to represent changes to instructions

    An ``Adjustment`` represents a change to program instructions, governed by one or more ``Adjustables``.
    This base class specifies the interface for all ``Adjustments``

    :param name: The name of the adjustment

    """

    def __init__(self, name):
        self.name = name
        self.adjustables = list()  #: A list of ``Adjustables``

    def get_initialization(self, progset: ProgramSet, instructions: ProgramInstructions) -> list:
        """
        Return initial values for ASD

        :param progset: The ``ProgramSet`` being used for the optimization
        :param instructions: The initial instructions
        :return: A list of initial values, one for each adjustable

        """
        return [x.initial_value for x in self.adjustables]

    def update_instructions(self, adjustable_values, instructions: ProgramInstructions):
        # adjustable_values contains the values for each adjustable in self.adjustables
        # at the current ASD iteration. This function updates the provided instructions in place
        return


class SpendingAdjustment(Adjustment):
    """
    Adjust program spending

    This adjustment class represents making a spending quantity adjustable. By default, the
    base class simply overwrites the spending value at a particular point in time
    A SpendingAdjustment has a separate Adjustable for each time reached (independently)

    :param prog_name: The code name of a program
    :param t: A single time, or list/array of times at which to make adjustments
    :param limit_type: Interpret ``lower`` and ``upper`` as absolute or relative limits (should be ``'abs'`` or ``'rel'``)
    :param lower: Lower bound (0 by default). A single value (used for all times) or a list/array the same length as ``t``
    :param upper: Upper bound (``np.inf`` by default). A single value (used for all times) or a list/array the same length as ``t``
    :param initial: Optionally specify the initial value, either as a scalar or list/array the same length as ``t``. If not specified,
                    the initial spend will be drawn from the program instructions, or the progset.
    """

    def __init__(self, prog_name, t, limit_type="abs", lower=0.0, upper=np.inf, initial=None):
        Adjustment.__init__(self, name=prog_name)
        self.prog_name = prog_name
        self.t = sc.promotetoarray(t)  # Time at which to apply the adjustment

        lower = sc.promotetolist(lower, keepnone=True)
        if len(lower) == 1:
            lower = lower * len(self.t)
        else:
            assert len(lower) == len(self.t), "If supplying lower bounds, you must either specify one, or one for every time point"

        upper = sc.promotetolist(upper, keepnone=True)
        if len(upper) == 1:
            upper = upper * len(self.t)
        else:
            assert len(upper) == len(self.t), "If supplying upper bounds, you must either specify one, or one for every time point"

        initial = sc.promotetolist(initial, keepnone=True)
        if len(initial) == 1:
            initial = initial * len(self.t)
        else:
            assert len(initial) == len(self.t), "If supplying initial values, you must either specify one, or one for every time point"

        self.adjustables = [Adjustable(prog_name, limit_type, lower_bound=lb, upper_bound=ub, initial_value=init) for lb, ub, init in zip(lower, upper, initial)]

    def update_instructions(self, adjustable_values, instructions: ProgramInstructions):
        # There is one Adjustable for each time point, so the adjustable_values
        # are a list of this same length, one value for each time point
        for i, t in enumerate(self.t):
            if self.prog_name not in instructions.alloc:
                instructions.alloc[self.prog_name] = TimeSeries(t=t, vals=adjustable_values[i])
            else:
                instructions.alloc[self.prog_name].insert(t, adjustable_values[i])

    def get_initialization(self, progset: ProgramSet, instructions: ProgramInstructions) -> list:
        """
        Return initial values for ASD

        The initial values correspond to either

        - The explicitly specified initial spend
        - The initial spend from the program set/instructions

        Note that the initial spend is NOT clipped to any bounds. This is because the initial spend is in turn used to compute
        relative spending constraints. If the initial spend is not consistent then an error will be subsequently raised
        at that point.

        :param progset: The ``ProgramSet`` being used for the optimization
        :param instructions: The initial instructions
        :return: A list of initial values, one for each adjustable

        """

        initialization = []
        for adjustable, t in zip(self.adjustables, self.t):
            if adjustable.initial_value is not None:
                initialization.append(adjustable.initial_value)
            else:
                alloc = progset.get_alloc(t, instructions)
                initialization.append(alloc[self.prog_name][0])  # The Adjustable's name corresponds to the name of the program being overwritten.
        return initialization


class StartTimeAdjustment(Adjustment):
    """
    Optimize program start year

    This is an example of an Adjustment that does not target a spending value

    """

    def __init__(self, name, lower, upper, initial):
        Adjustment.__init__(self, name=name)
        self.adjustables = [Adjustable("start_year", limit_type="abs", lower=lower, upper=upper, initial=initial)]

    def update_instructions(self, adjustable_values, instructions: ProgramInstructions):
        instructions.start_year = adjustable_values[0]

    def get_initialization(self, progset, instructions: ProgramInstructions):
        if self.initial_value:
            return self.initial_value
        else:
            return instructions.start_year


class ExponentialSpendingAdjustment(Adjustment):
    """
    Parametric overwrite example

    This is an example of an Adjustment that uses a function of several variables to
    compute time-dependent spending.

    """

    # Example of a parametric time-varying budget where multiple parameters
    # are mapped to spending via a function

    def __init__(self, prog_name, t, t_0, t_end, p1, a1, a2):
        Adjustment.__init__(self, name=prog_name)
        self.prog_name = prog_name
        self.t = t  # Vector of time values instructions are updated at
        self.t_0 = t_0  # non-adjustable parameter
        self.t_end = t_end  # non_adjustable parameter
        self.adjustables = [Adjustable("p1", initial_value=p1), Adjustable("a1", initial_value=a1), Adjustable("a2", initial_value=a2)]  # Would need to specify initial values and limits for these parameters

    # Note - we can use get_initialization from the base class because all of the Adjustables
    # have explicit initial values. Note that it's essential to provide explicit initial values
    # for any Adjustable that does not explicitly appear in the instructions. Although in theory,
    # it would be possible to write a `get_initialization` function that fits the Adjustables
    # to the initial spending...

    def update_instructions(self, adjustable_values, instructions: ProgramInstructions):
        # p1*exp(a1*(t-t_0))*exp(b1*(t-t_end))
        instructions.alloc[self.prog_name][self.t] = adjustable_values[0] * np.exp(adjustable_values[1] * (self.t - self.t_0)) * np.exp(adjustable_values[2] * (self.t - self.t_end))


class SpendingPackageAdjustment(Adjustment):
    """
    Adjustment to set total spending on several programs

    This adjustment can be used when it is important to constrain the sum of spending on multiple programs. For example,
    a package could be defined for 'MDR short' and 'MDR standard' standard programs, where the adjustments would be the
    total spend on MDR, and the fraction spent on MDR short.
    """

    def __init__(self, package_name: str, t: float, prog_names: list, initial_spends: np.array, min_props: list = None, max_props: list = None, min_total_spend: float = None, max_total_spend: float = None, fix_props: bool = False):
        """

        By default, the total spend on the package is constrained. In practice, it would usual to either
            - Fix the proportions of the programs within the package
            OR
            - Fix the total spend on the package

        Although a less common usage would be to constrain the ratio of the programs within the
        package (e.g., program B must have no more than half the funding of program A).

        Essentially, if none of `min_props`, `max_props`, `min_total_spend` or `max_total_spend` are
        specified, the SpendingPackageAdjustment should function the same as simply optimizing each
        intervention independently.

        :param package_name: The name of this spending package
        :param t: Year in which to apply spending adjustments
        :param prog_names: List of program names to include in this package
        :param initial_spends: Array of spending values for each program, same length as `prog_names`
        :param min_props: (optional) List of minimum proportion to spend on each program, same length as ``prog_names``. Default: 0 for all programs
        :param max_props: (optional) List of maximum proportion to spend on each program, same length as ``prog_names``. Default: 1 for all programs
        :param min_total_spend: (optional) Minimum total spend. Default: equal to the sum of the initial spends
        :param max_total_spend: (optional) Maximum total spend. Default: equal to the sum of the initial spends
        :param fix_props: (optional) If True, do not allow initial spending proportion to change
        """

        self.name = package_name
        self.prog_name = prog_names
        self.t = t
        self.min_props = np.zeros(len(prog_names)) if min_props is None else sc.promotetoarray(min_props)
        self.max_props = np.ones(len(prog_names)) if max_props is None else sc.promotetoarray(max_props)
        self.initial_spends = sc.promotetoarray(initial_spends)

        initial_total = self.initial_spends.sum()

        assert self.min_props.sum() <= 1, "Constraining fractions of an intervention package where the minimum total is >1 is impossible"
        assert self.max_props.sum() >= 1, "Constraining fractions of an intervention package where the maximum total is <1 is impossible"

        self.adjustables = []

        if initial_total == 0.0:
            initial_props = np.array([1.0 / len(initial_spends) for _ in initial_spends])  # if there was no spending on the package initially, assume equal proportions for all components
        else:
            initial_props = self.initial_spends / initial_total

        if not fix_props:
            for program, initial_prop, lb, ub in zip(prog_names, initial_props, self.min_props, self.max_props):
                assert initial_prop >= lb, f"SpendingProgramAdjustment: Initial spend on {program} is a smaller proportion ({initial_prop}) of the package than requested minimum proportion ({lb})"
                assert initial_prop <= ub, f"SpendingProgramAdjustment: Initial spend on {program} is a larger proportion ({initial_prop}) of the package than requested maximum proportion ({ub})"
                # set to the current proportion - note that the overall fractions won't end up being constrained so will have to be rescaled
                # Similarly, the upper and lower bounds here will still need to be enforced later after scaling when updating instructions
                self.adjustables.append(Adjustable("frac_" + program, initial_value=initial_prop, lower_bound=lb, upper_bound=ub))
        else:
            # If the initial proportions are not to be changed, then the min_props and max_props arguments have no effect. Warn the
            # user if they have been provided and are going to be ignored (but not with logger.warning since this might be expected)
            if self.min_props is not None:
                logger.info("Minimum proportions are being ignored because the `fix_props` flag has been set")
            if self.max_props is not None:
                logger.info("Maximum proportions are being ignored because the `fix_props` flag has been set")

        if min_total_spend is not None:
            assert initial_total >= min_total_spend, "SpendingProgramAdjustment: Initial total spend is less than minimum constraint"
        else:
            min_total_spend = initial_total

        if max_total_spend is not None:
            assert initial_total <= max_total_spend, "SpendingProgramAdjustment: Initial total spend is greater than maximum constraint"
        else:
            max_total_spend = initial_total

        # total spending can be adjusted as a single adjustable
        if not (min_total_spend == initial_total and min_total_spend == max_total_spend):
            self.adjustables.append(Adjustable("package_spend", initial_value=initial_total, lower_bound=min_total_spend, upper_bound=max_total_spend))

    @property
    def adjust_total_spend(self) -> bool:
        """
        Returns True if the total spend on this package can be optimized
        :return:
        """
        if self.adjustables:
            return self.adjustables[-1].name == "package_spend"
        else:
            return False

    @property
    def fix_props(self) -> bool:
        if len(self.adjustables) == 0 or self.adjustables[0].name == "package_spend":
            return True
        else:
            return False

    def update_instructions(self, adjustable_values, instructions):

        adjustable_values = sc.promotetoarray(adjustable_values)

        if self.fix_props:
            if self.initial_spends.sum() == 0:
                fracs = np.array([1.0 / len(self.initial_spends) for _ in self.initial_spends])  # evenly distributed fractions though it seems unlikely that there would be constrained fractions of zero spending
            else:
                fracs = self.initial_spends / self.initial_spends.sum()
        elif self.adjust_total_spend:
            fracs = adjustable_values[:-1]
        else:
            fracs = adjustable_values

        # Scale the fractions such that they add to 1
        fracs = constrain_sum_bounded(fracs, 1, self.min_props, self.max_props)

        # now that the fractions sum to 1 and don't violate any constraints, update the spending on each program
        if self.adjust_total_spend:
            total_spending = adjustable_values[-1]  # last on the list
        else:
            total_spending = self.initial_spends.sum()

        for pn, prog_name in enumerate(self.prog_name):
            this_prog_spend = fracs[pn] * total_spending
            if prog_name not in instructions.alloc:
                instructions.alloc[prog_name] = TimeSeries(t=self.t, vals=this_prog_spend)
            else:
                instructions.alloc[prog_name].insert(t=self.t, v=this_prog_spend)

    def get_total_spend(self, instructions):
        return sum([instructions.alloc[prog_name].get(self.t) for prog_name in self.prog_name])

    def set_total_spend(self, instructions, total_spend):
        if self.get_total_spend(instructions) > 0:
            spend_factor = total_spend / self.get_total_spend(instructions)
        else:
            spend_factor = 0.0  # if total spending is zero, spending on each program must be zero?
        for prog in self.prog_name:
            ts = instructions.alloc[prog]
            ts.insert(t=self.t, v=ts.get(self.t) * spend_factor)


class PairedLinearSpendingAdjustment(Adjustment):
    """
    Parametric overwrite with multiple programs

    This example Adjustment demonstrates a parametric time-varying budget reaching more than one
    program. A single adjustable corresponding to the rate of change simultaneously acts on
    two programs in opposite directions

    """

    def __init__(self, prog_names, t):
        Adjustment.__init__(self, name=None)
        assert len(prog_names) == 2, "PairedLinearSpendingAdjustment needs exactly two program names"
        self.prog_name = prog_names  # Store names in the attribute 'prog_name' so that TotalSpendConstraint picks this up
        self.t = t  # [t_start,t_stop] for when to start/stop ramping
        self.adjustables = [Adjustable("ramp", initial_value=0.0)]

    def update_instructions(self, adjustable_values, instructions: ProgramInstructions):

        gradient = adjustable_values[0]
        tspan = self.t[1] - self.t[0]

        if gradient < 0:  # We are transferring money from program 2 to program 1
            available = -instructions.alloc[self.prog_name[1]].get(self.t[0])
            max_gradient = float(available) / tspan
            if gradient < max_gradient:
                gradient = max_gradient
        else:
            available = instructions.alloc[self.prog_name[0]].get(self.t[0])
            max_gradient = float(available) / tspan
            if gradient > max_gradient:
                gradient = max_gradient

        funding_change = gradient * tspan  # Amount transferred from program 1 to program 2

        # This does not change the amount of funds allocated in the initial year
        for prog in self.prog_name:
            instructions.alloc[prog].insert(self.t[0], instructions.alloc[prog].interpolate(self.t[0])[0])
            instructions.alloc[prog].remove_between(self.t)

        instructions.alloc[self.prog_name[0]].insert(self.t[1], instructions.alloc[self.prog_name[0]].get(self.t[0]) - funding_change)
        instructions.alloc[self.prog_name[1]].insert(self.t[1], instructions.alloc[self.prog_name[1]].get(self.t[0]) + funding_change)


class Measurable:
    """
    Optimization objective

    A ``Measurable`` is a class that returns an objective value based on a simulated
    ``Model`` object. It takes in a ``Model`` and returns a scalar value. Often, an optimization
    may contain multiple ``Measurable`` objects, and the objective value returned by each of them
    is summed together.

    :param measurable_name: The base measurable class accepts the name of a program (for spending) or a quantity supported by ``Population.get_variable()``
    :param t: Single year, or a list of two start/stop years. If specifying a single year, that year must appear in the simulation output.
              The quantity will be summed over all simulation time points
    :param pop_names: The base `Measurable` class takes in the names of the populations to use. If multiple populations are provided, the objective will be added across the named populations
    :param weight: The weight factor multiplies the quantity

    """

    def __init__(self, measurable_name, t, pop_names=None, weight=1.0):
        self.measurable_name = measurable_name
        self.t = sc.promotetoarray(t)
        assert len(self.t) <= 2, "Measurable time must either be a year, or the `[low,high)` values defining a period of time"
        self.weight = weight
        self.pop_names = pop_names

    def eval(self, model, baseline):
        # This is the main interface with the optimization code - this function gets called
        # to return the transformed, weighted objective value from this Measurable for use in ASD
        # Only overload this if you want to customize the transformation and weighting
        return self.weight * self.get_objective_val(model, baseline)

    def get_baseline(self, model):
        """
        Return cached baseline values

        Similar to ``get_hard_constraint``, sometimes a relative ``Measurable`` might be desired e.g. 'Reduce deaths by at least 50%'.
        In that case, we need to perform a procedure similar to getting a hard constraint, where the ``Measurable`` receives an initial
        ``Model`` object and extracts baseline data for subsequent use in ``get_objective_val``.

        Thus, the output of this function is paired to its usage in ``get_objective_val``.

        :param model:
        :return: The value to pass back to the ``Measurable`` during optimization
        """

        return None

    def get_objective_val(self, model: Model, baseline) -> float:
        """
        Return objective value

        This method should return the _unweighted_ objective value. Note that further transformation may occur

        :param model: A ``Model`` object after integration
        :param baseline: The baseline variable returned by this ``Measurable`` at the start of optimization
        :return: A scalar objective value

        """

        # This returns the base objective value, prior to any function transformation
        # or weighting. The function transformation and weighting are handled by this base
        # class - derived classes may wish to not expose things like the function mapping
        # in their constructors, if that wouldn't be appropriate. But otherwise, they can
        # inherit this behaviour for free. So derived classes should overload this method
        #
        # The base class has the default behaviour that the 'measurable name' is a model variable
        if len(self.t) == 1:
            t_filter = model.t == self.t  # boolean vector for whether to use the time point or not. This could be relaxed using interpolation if needed, but safer not to unless essential
        else:
            t_filter = (model.t >= self.t[0]) & (model.t < self.t[1])  # Don't include upper bound, so [2018,2019] will include exactly one year

        if self.measurable_name in model.progset.programs:
            alloc = model.progset.get_alloc(model.t, model.program_instructions)
            val = np.sum(alloc[self.measurable_name][t_filter])
        else:  # If the measurable is a model output...
            val = 0.0
            matched = False  # Flag whether any variables were found
            for pop in model.pops:
                if not self.pop_names:
                    # If no pops were provided, then iterate over all pops but skip those where the measureable is not defined
                    # Use this approach rather than checking the pop type in the framework because user could be optimizating
                    # flow rates or transitions that don't appear in the framework
                    try:
                        vars = pop.get_variable(self.measurable_name)
                        matched = True
                    except NotFoundError:
                        continue
                elif pop not in self.pop_names:
                    continue
                else:
                    vars = pop.get_variable(self.measurable_name)  # If variable is missing and the pop was explicitly defined, raise the error
                    matched = True

                for var in vars:
                    if isinstance(var, Link):
                        val += np.sum(var.vals[t_filter] / var.dt)  # Annualize link values - usually this won't make a difference, but it matters if the user mixes Links with something else in the objective
                    else:
                        val += np.sum(var.vals[t_filter])

            if not matched:
                # Raise an error if the measureable was not found in any populations
                raise Exception('"%s" not found in any populations' % (self.measurable_name))

        return val


class MinimizeMeasurable(Measurable):
    # Syntactic sugar for a measurable that minimizes its quantity
    def __init__(self, measurable_name, t, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=1, pop_names=pop_names)


class MaximizeMeasurable(Measurable):
    # Syntactic sugar for a measurable that maximizes its quantity
    def __init__(self, measurable_name, t, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=-1, pop_names=pop_names)


class AtMostMeasurable(Measurable):
    """
    Enforce quantity is below a value

    This Measurable imposes a penalty if the quantity is larger than some threshold
    The initial points should be 'valid' in the sense that the quantity starts out
    below the threshold (and during optimization it will never be allowed to cross
    the threshold).

    Typically, this Measurable would be used in conjunction with other measurables -
    for example, optimizing one quantity while ensuring another quantity does not
    cross a threshold.

    The measurable returns ``np.inf`` if the condition is violated, and ``0.0`` otherwise.

    """

    def __init__(self, measurable_name, t, threshold, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        self.weight = 1.0
        self.threshold = threshold

    def get_objective_val(self, model, baseline):
        val = Measurable.get_objective_val(self, model, baseline)
        return np.inf if val > self.threshold else 0.0

    def __repr__(self):
        return "AtMostMeasurable(%s < %f)" % (self.measurable_name, self.threshold)


class AtLeastMeasurable(Measurable):
    """
    Enforce quantity exceeds a value

    This Measurable imposes a penalty if the quantity is smaller than some threshold
    The initial points should be 'valid' in the sense that the quantity starts out
    above the threshold (and during optimization it will never be allowed to cross
    the threshold)

    Typically, this Measurable would be used in money minimization in conjunction
    with measurables that aim to minimize spending.

    The measurable returns ``np.inf`` if the condition is violated, and ``0.0`` otherwise.

    """

    def __init__(self, measurable_name, t, threshold, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        self.weight = 1.0
        self.threshold = threshold

    def get_objective_val(self, model, baseline):
        val = Measurable.get_objective_val(self, model, baseline)
        return np.inf if val < self.threshold else 0.0

    def __repr__(self):
        return "AtLeastMeasurable(%s > %f)" % (self.measurable_name, self.threshold)


class IncreaseByMeasurable(Measurable):
    """
    Increase quantity by percentage

    This Measurable stores the value of a quantity using the original instructions.
    It then requires that there is a minimum increase in the value of the quantity
    during optimization. For example

    >>> IncreaseByMeasurable('alive',2030,0.05)

    This Measurable would correspond to an increase of 5% in the number of people
    alive in 2030.

    The measurable returns ``np.inf`` if the condition is violated, and ``0.0`` otherwise.

    :param measurable_name: The base measurable class accepts the name of a program (for spending) or a quantity supported by ``Population.get_variable()``
    :param t: Single year, or a list of two start/stop years. If specifying a single year, that year must appear in the simulation output.
              The quantity will be summed over all simulation time points
    :param increase: The amount by which to increase the measurable (e.g. 0.05 for a 5% increase). Use ``target_type='abs'`` to specify an absolute increase
    :param pop_names: The base `Measurable` class takes in the names of the populations to use. If multiple populations are provided, the objective will be added across the named populations
    :param target_type: Specify fractional 'frac' or absolute 'abs' increase (default is fractional)

    """

    def __init__(self, measurable_name, t, increase, pop_names=None, target_type="frac"):

        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        assert increase >= 0, "Cannot set negative increase"
        self.weight = 1.0
        self.increase = increase  # Required increase
        self.target_type = target_type

    def get_baseline(self, model) -> float:
        return Measurable.get_objective_val(self, model, None)  # Get the baseline value using the underlying Measurable

    def get_objective_val(self, model: Model, baseline: float) -> float:
        val = Measurable.get_objective_val(self, model, None)
        if self.target_type == "frac":
            return np.inf if (val / baseline) < (1 + self.increase) else 0.0
        elif self.target_type == "abs":
            return np.inf if val < (baseline + self.increase) else 0.0
        else:
            raise Exception("Unknown target type")


class DecreaseByMeasurable(Measurable):
    """
    Decrease quantity by percentage

    This Measurable stores the value of a quantity using the original instructions.
    It then requires that there is a minimum increase in the value of the quantity
    during optimization. For example

    >>> DecreaseByMeasurable('deaths',2030,0.05)

    This Measurable would correspond to an decrease of 5% in the number of deaths in 2030.

    The measurable returns ``np.inf`` if the condition is violated, and ``0.0`` otherwise.

    :param measurable_name: The base measurable class accepts the name of a program (for spending) or a quantity supported by ``Population.get_variable()``
    :param t: Single year, or a list of two start/stop years. If specifying a single year, that year must appear in the simulation output.
              The quantity will be summed over all simulation time points
    :param decrease: The amount by which to decrease the measurable (e.g. 0.05 for a 5% decrease). Use ``target_type='abs'`` to specify an absolute decrease
    :param pop_names: The base `Measurable` class takes in the names of the populations to use. If multiple populations are provided, the objective will be added across the named populations
    :param target_type: Specify fractional 'frac' or absolute 'abs' decrease (default is fractional)

    """

    def __init__(self, measurable_name, t, decrease, pop_names=None, target_type="frac"):

        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        self.weight = 1.0
        assert decrease >= 0, "Set positive value for magnitude of decrease e.g. 0.05 for a 5%% decrease"
        assert target_type == "abs" or decrease <= 1, "Cannot decrease by more than 100%% - fractional decrease should be a value between 0 and 1"
        self.decrease = decrease
        self.target_type = target_type

    def get_baseline(self, model) -> float:
        return Measurable.get_objective_val(self, model, None)  # Get the baseline value using the underlying Measurable

    def get_objective_val(self, model: Model, baseline: float) -> float:
        val = Measurable.get_objective_val(self, model, None)
        if self.target_type == "frac":
            return np.inf if (val / baseline) > (1 - self.decrease) else 0.0
        elif self.target_type == "abs":
            return np.inf if val > (baseline - self.decrease) else 0.0
        else:
            raise Exception("Unknown target type")


class MaximizeCascadeStage(Measurable):
    # This Measurable will maximize the number of people in a cascade stage, whatever that stage is
    # e.g. `measurable = MaximizeCascadeStage('main',2020)
    # If multiple years are provided, they will be summed over
    # Could add option to weight specific years later on, if desired

    def __init__(self, cascade_name, t, pop_names="all", weight=1.0, cascade_stage=-1):
        # Instantiate thpop_names can be a single pop name (including all), or a list of pop names
        # aggregations are supported by setting pop_names to a dict e.g.
        # pop_names = {'foo':['0-4','5-14']}
        #
        # The cascade name is passed to `get_cascade_vals` so any cascade can be passed in
        # Usually this would be 'None' (fallback cascade, or first cascade) or the name of a
        # cascade
        #
        # - t : Can be a single time or array of times (as supposed by get_cascade_vals)
        # - cascade_stage specifies which stage(s) to include. It can be
        #   - An integer indexing the `cascade_vals` dict e.g. -1 is the last stage
        #   - The name of a cascade stage
        #   - A list of integers and/or names to include multiple stages

        Measurable.__init__(self, cascade_name, t=t, weight=-weight, pop_names=pop_names)
        if not isinstance(cascade_stage, list):
            cascade_stage = [cascade_stage]
        self.cascade_stage = cascade_stage
        if not isinstance(self.pop_names, list):
            self.pop_names = [self.pop_names]

    def get_objective_val(self, model, baseline):
        result = Result(model=model)
        val = 0
        for pop_name in self.pop_names:
            cascade_vals = get_cascade_vals(result, self.measurable_name, pop_name, self.t)
            for stage in self.cascade_stage:  # Loop over included stages
                val += np.sum(cascade_vals[0][stage])  # Add the values from the stage, summed over time
        return val


class MaximizeCascadeConversionRate(Measurable):
    """
    Maximize overall conversion rate

    Maximize conversion summed over all cascade stages

    :param cascade_name: The name of one of the cascades in the Framework
    :param t: A single time value e.g. 2020
    :param pop_names: A single pop name (including 'all'), a list of populations,
                  or a dict/list of dicts, each with a single aggregation e.g. ``{'foo':['0-4','5-14']}``
    :param weight: Weighting factor for this Measurable in the overall objective function

    """

    def __init__(self, cascade_name, t: float, pop_names="all", weight=1.0):
        Measurable.__init__(self, cascade_name, t=t, weight=-weight, pop_names=pop_names)
        if not isinstance(self.pop_names, list):
            self.pop_names = [self.pop_names]

    def get_objective_val(self, model, baseline):
        if self.t < model.t[0] or self.t > model.t[-1]:
            raise Exception("Measurable year for optimization (%d) is outside the simulation range (%d-%d)" % (self.t, model.t[0], model.t[-1]))
        result = Result(model=model)
        val = 0
        for pop_name in self.pop_names:
            cascade_vals = get_cascade_vals(result, self.measurable_name, pop_name, self.t)[0]
            cascade_array = np.hstack(cascade_vals.values())
            conversion = cascade_array[1:] / cascade_array[0:-1]
            val += np.sum(conversion)
        return val


class Constraint:
    """
    Store conditions to satisfy during optimization

    A Constraint represents a condition that must be satisfied by the Instructions
    after the cumulative effect of all adjustments. The Instructions are rescaled to
    satisfy the constraint directly (rather than changing the value of the Adjustables)
    although this distinction really only matters in the context of parametric spending.

    """

    def get_hard_constraint(self, optimization, instructions: ProgramInstructions):
        """
        Return hard constraint from initial instructions

        Often constraints can be specified relative to the initial conditions. For example,
        fixing total spend regardless of what the total spend is in the initial instructions.
        Therefore, during ``constrain_instructions``, it is necessary to examine properties
        from the initial instructions in order to perform the constraining.

        This method is called at the very start of optimization, passing in the initial instructions.
        It then returns an arbitrary value that is passed back to the instance's ``constrain_instructions``
        during optimization. For example, consider the total spending constraint

        - ``get_hard_constraint`` would extract the total spend from the initial instructions
        - This value is passed to ``constrain_instructions`` where it is used to rescale spending

        Because subclasses implement both ``get_hard_constraint`` and ``constrain_instructions``
        no assumptions need to be made about the value returned by this method - it simply needs
        to be paired to ``constrain_instructions``.

        :param optimization: An ``Optimization``
        :param instructions: A set of initial instructions to extract absolute constraints from
        :return: Arbitrary variable that will be passed back during ``constrain_instructions``
        """

        return

    def constrain_instructions(self, instructions: ProgramInstructions, hard_constraints, optimization) -> float:
        """
        Apply constraint to instructions

        Constrains the instructions, returns a metric penalizing the constraint
        If there is no penalty associated with adjusting (perhaps if all of the Adjustments are
        parametric?) then this would be 0.0. The penalty represents in some sense the quality
        of the constraint. For example, the default ``TotalSpendConstraint`` rescales spending
        such that the total spend matches a target value. The penalty reflects the distance between
        the requested spend and the constrained spend, so it is desirable to minimize it.

        If it is not possible to constrain the instructions, raise ``FailedConstraint``.

        :param instructions: The ``ProgramInstructions`` instance to constrain (in place)
        :param hard_constraints: The hard constraint returned by ``get_hard_constraint``
        :param optimization: The parent optimization, in case it is needed
        :return: A numeric penalty value. Return `np.inf` if constraint penalty could not be computed
        :raises: :class:`FailedConstraint` if the instructions could not be constrained

        """

        return 0.0


class TotalSpendConstraint(Constraint):
    """
    Fix total spending

    This class implements a constraint on the total spend at every time point when a program
    is optimizable. A program is considered optimizable if an Adjustment reaches that program
    at the specified time. Spending is constrained independently at all times when any program
    is adjustable.

    The ``initial_total_spend`` argument allows the total spending in a particular year to be explicitly specified
    rather than drawn from the initial allocation. This can be useful when using parametric programs where
    the adjustables do not directly correspond to spending value. If the total spend is not provided, it will
    automatically be computed from the first ASD step. Note that it is computed based on the initial instructions
    *after* the initial ASD values have been applied. This is because relative constraints on all adjustables are
    interpreted relative to the initial value of the adjustable, since not all adjustables map directly to values
    in the instructions. Since the adjustable hard upper and lower bounds are used as part of the constraint, for
    consistency, the total spend constraint itself is drawn from the same set of instructions (i.e. after the initial
    value has been applied). In cases where this is not the desired behaviour, the cause would likely be that the
    default value does not agree with a known desired total spend value. In that case, the desired total spend should
    simply be specified here explicitly as an absolute value.

    This constraint can also be set to only apply in certain years.

    The ``budget_factor`` multiplies the total spend at the time the ``hard_constraint`` is assigned
    Typically this is to scale up the available spending when that spending is being drawn from
    the instructions/progset (otherwise the budget_factor could already be part of the specified total spend)

    Note that if no times are specified, the budget factor should be a scalar but no explicit
    spending values can be specified. This is because in the case where different programs are
    optimized in different years, an explicit total spending constraint applying to all
    times is unlikely to be a sensible choice (so we just ask the user to specify the time as well).

    :param total_spend: A list of spending amounts the same size as ``t`` (can contain Nones), or ``None``.
                        For times in which the total spend is ``None``, it will be automatically set to the sum of
                        spending on optimizable programs in the corresponding year
    :param t: A time, or list of times, at which to apply the total spending constraint. If None, it will automatically be set to all years in which spending adjustments are being made
    :param budget_factor: The budget factor multiplies whatever the ``initial_total_spend`` is. This can either be a single value, or a year specific value

    """

    def __init__(self, total_spend=None, t=None, budget_factor=1.0):
        self.total_spend = sc.promotetoarray(total_spend) if total_spend is not None else ()
        self.t = sc.promotetoarray(t) if t is not None else ()
        self.budget_factor = sc.promotetoarray(budget_factor)

        if t is None:
            assert total_spend is None, "If no times are specified, no total spend values can be specified either"
            assert len(self.budget_factor) == 1, "If no times are specified, the budget factor must be scalar"

        if t is not None and total_spend is not None:
            assert len(self.total_spend) == len(self.t), "If specifying both the times and values for the total spending constraint, their lengths must be the same"
        if len(self.budget_factor) > 1:
            assert len(self.budget_factor) == len(self.t), "If specifying multiple budget factors, you must also specify the years in which they are used"

    def get_hard_constraint(self, optimization, instructions: ProgramInstructions) -> dict:
        """
        Return hard constraint dictionary

        :param optimization: ``Optimization`` instance
        :param instructions: Initial ``ProgramInstructions``
        :return:
        """

        # First, at each time point where a program overwrite exists, we need to store
        # the names of all of the programs being overwritten
        # e.g.
        # hard_constraints['programs'][2020] = ['Program 1','Program 2']
        # hard_constraints['programs'][2030] = ['Program 2','Program 3']
        # By convention, an Adjustable affecting a program should store the program's name in the
        # `prog_name` attribute. If a program reaches multiple programs, then `prog_name` will be a list

        # First, we make a dictionary storing the years in which any adjustments are made
        # as well as the upper and lower bounds in that year. This is only done for spending
        # adjustments (i.e. those where `adjustment.prog_name` is defined)
        _package_adjustments = set()  # Check for duplicate names
        hard_constraints = {}
        hard_constraints["programs"] = defaultdict(set)  # It's a set so that it will work properly if multiple Adjustments reach the same parameter at the same time. However, this would a bad idea and nobody should do this!
        for adjustment in optimization.adjustments:
            if hasattr(adjustment, "prog_name"):
                for t in sc.promotetoarray(adjustment.t):
                    if isinstance(adjustment, SpendingPackageAdjustment):
                        if adjustment.adjust_total_spend and adjustment.t == t:
                            hard_constraints["programs"][t].add(adjustment.name)
                            if adjustment.name in _package_adjustments:
                                raise Exception("More than one `SpendingPackageAdjustment` in the optimization has the same name - not compatible with total spend constraint")
                            else:
                                _package_adjustments.add(adjustment.name)
                    elif isinstance(adjustment.prog_name, list):
                        hard_constraints["programs"][t].update(adjustment.prog_name)
                    else:
                        hard_constraints["programs"][t].add(adjustment.prog_name)

        if len(self.t):
            # Check that every explictly specified time has
            # corresponding spending adjustments available for use
            missing_times = set(self.t) - set(hard_constraints["programs"].keys())
            if missing_times:
                raise Exception("Total spending constraint was specified in %s but the optimization does not have any adjustments at those times" % (missing_times))

        # Now we have a set of times and programs for which we need to get total spend, and
        # also which programs should be included in the total for that year
        #
        # hard_constraints['initial_total_spend'][2020] = 300
        # hard_constraints['initial_total_spend'][2030] = 400
        hard_constraints["initial_total_spend"] = {}
        for t, progs in hard_constraints["programs"].items():
            # For every time at which programs are optimizable...
            if len(self.t) and t not in self.t:
                # If we are not wanting to constrain spending in this year, then continue
                continue
            elif len(self.t):
                idx = np.where(self.t == t)[0][0]  # This is the index for the constraint year
            else:
                idx = None

            if not len(self.total_spend) or self.total_spend[idx] is None:
                # Get the total spend from the allocation in this year
                total_spend = 0.0
                for prog in progs:
                    if prog in instructions.alloc:
                        total_spend += instructions.alloc[prog].get(t)
                    else:
                        # It's a SpendingPackageAdjustment name
                        adj = optimization.get_adjustment(prog)
                        if adj.adjust_total_spend:
                            total_spend += adj.initial_spends.sum()
            else:
                total_spend = self.total_spend[idx]

            # Lastly, apply the budget factor
            if len(self.budget_factor) == 1:
                hard_constraints["initial_total_spend"][t] = total_spend * self.budget_factor
            else:
                hard_constraints["initial_total_spend"][t] = total_spend * self.budget_factor[idx]

        # Finally, for each adjustable, we need to store its upper and lower bounds
        # _in the year that the adjustment is being made_
        #
        # This is because in the case where we have an multiple adjustables targeting the same Program at different
        # time points, the upper/lower bounds might be different. For example, suppose we have spending on
        # Program 1 specified somehow as 2020=$20 and 2030=$30. Then suppose we add an Adjustment for Program 1 in
        # 2020 *and* 2030 with relative bounds of [0.5,2.0]. Then, in 2020 spending on Program 1 is limited to
        # $10-40 while in 2030 it's limited to $15-$60. Thus the spending constraint could also be time-varying.
        # In the case of a parametric overwrite, there will be no direct spending bounds based on the Adjustable. Then,
        # the spending should only be restricted to a positive value.

        hard_constraints["bounds"] = dict()

        for t, progs in hard_constraints["programs"].items():  # For each time point being constrained, and for each program
            if t not in hard_constraints["initial_total_spend"]:
                # If the time is not one where the total spending constraint is being applied, then
                # just skip it
                continue

            hard_constraints["bounds"][t] = dict()
            # If there is an Adjustable that reaches this Program in the appropriate year:

            # Keep track of the absolute lower and upper bounds on spending permitted by the program constraints
            minimum_spend = 0.0
            maximum_spend = 0.0

            for adjustment in optimization.adjustments:
                if isinstance(adjustment, SpendingPackageAdjustment):
                    if adjustment.adjust_total_spend and adjustment.t == t:
                        hard_constraints["bounds"][t][adjustment.name] = (adjustment.adjustables[-1].lower_bound, adjustment.adjustables[-1].upper_bound)
                        minimum_spend += adjustment.adjustables[-1].lower_bound
                        maximum_spend += adjustment.adjustables[-1].upper_bound
                elif hasattr(adjustment, "prog_name") and np.array([prog in progs for prog in sc.promotetolist(adjustment.prog_name)]).all() and t in adjustment.t:
                    if isinstance(adjustment, SpendingAdjustment):
                        idx = np.where(adjustment.t == t)[0][0]  # If it is a SpendingAdjustment then set bounds from the appropriate Adjustable
                        adjustable = adjustment.adjustables[idx]
                        hard_constraints["bounds"][t][adjustment.prog_name] = adjustable.get_hard_bounds(instructions.alloc[adjustment.prog_name].get(t))  # The instructions should already have the initial spend on this program inserted. This may be inconsistent if multiple Adjustments reach the same program...!
                    else:
                        for prog in sc.promotetolist(adjustment.prog_name):
                            hard_constraints["bounds"][t][prog] = (0.0, np.inf)  # If the Adjustment reaches spending but is not a SpendingAdjustment then do not constrain the alloc

                    for prog in sc.promotetolist(adjustment.prog_name):
                        minimum_spend += hard_constraints["bounds"][t][prog][0]
                        maximum_spend += hard_constraints["bounds"][t][prog][1]

            if minimum_spend > hard_constraints["initial_total_spend"][t]:
                raise UnresolvableConstraint("The total spend in %.2f is constrained to %.2f but the individual programs have a total minimum spend of %.2f which is impossible to satisfy. Please either raise the total spending, or lower the minimum spend on one or more programs" % (t, hard_constraints["initial_total_spend"][t], minimum_spend))

            if maximum_spend < hard_constraints["initial_total_spend"][t]:
                raise UnresolvableConstraint("The total spend in %.2f is constrained to %.2f but the individual programs have a total maximum spend of %.2f which is impossible to satisfy. Please either lower the total spending, or raise the maximum spend on one or more programs" % (t, hard_constraints["initial_total_spend"][t], maximum_spend))

        return hard_constraints

    def constrain_instructions(self, instructions: ProgramInstructions, hard_constraints: dict, optimization) -> float:
        """
        Apply total spend constraint

        :param instructions: The ``ProgramInstructions`` instance to constrain
        :param hard_constraints: Dictionary of hard constraints
        :param optimization: The parent optimization, in case it's needed
        :return: Distance-like difference between initial spending and constrained spending, `np.inf` if constraint failed

        """

        penalty = 0.0

        for t, total_spend in hard_constraints["initial_total_spend"].items():

            total_spend = sc.promotetoarray(total_spend).ravel()[0]  # Make sure total spend is a scalar
            x0 = sc.odict()  # Order matters here
            lb = []
            ub = []
            progs = hard_constraints["programs"][t]  # Programs eligible for constraining at this time

            for prog in progs:
                if prog not in instructions.alloc:
                    # The program name corresponds to the name of a SpendingPackageAdjustment
                    adj = optimization.get_adjustment(prog)  # Find the adjustment. Note that we find the adjustment by name, therefore SpendingPackageAdjustment names must be unique
                    x0[prog] = adj.get_total_spend(instructions)
                else:
                    x0[prog] = instructions.alloc[prog].get(t)

                low, high = hard_constraints["bounds"][t][prog]
                lb.append(low)
                ub.append(high)

            x0_array = np.array(x0.values()).ravel()
            lb = np.array(lb)
            ub = np.array(ub)

            x1_array = constrain_sum_bounded(x0_array, total_spend, lb, ub)  # scaled spending that optimally satisfies constraints

            penalty += total_spend * np.linalg.norm(x1_array - x0_array)  # Penalty is the distance between the unconstrained budget and the constrained budget
            for name, val in zip(x0.keys(), x1_array):
                if name in instructions.alloc:
                    instructions.alloc[name].insert(t, val)
                else:
                    adj = optimization.get_adjustment(name)
                    adj.set_total_spend(instructions, val)

        return penalty


class Optimization(NamedItem):
    """
    Instructions on how to perform an optimization

    The Optimization object stores the information that defines an optimization operation.
    Optimization can be thought of as a function mapping one set of program instructions
    to another set of program instructions. The parameters of that function are stored in the
    Optimization object, and amount to

    - A definition of optimality
    - A specification of allowed changes to the program instructions
    - Any additional information required by a particular optimization algorithm e.g. ASD

    :param name:
    :param adjustments: An `Adjustment` or list of `Adjustment` objects
    :param measurables: A `Measurable` or list of `Measurable` objects
    :param constraints: Optionally provide a `Constraint` or list of `Constraint` objects
    :param maxtime: Optionally specify maximum ASD time
    :param maxiters: Optionally specify maximum number of ASD iterations or hyperopt evaluations
    :param method: One of ['asd','pso','hyperopt'] to use
                        - asd (to use normal ASD)
                        - pso (to use particle swarm optimization from pyswarm)
                        - hyperopt (to use hyperopt's Bayesian optimization function)

    """

    def __init__(self, name=None, adjustments=None, measurables=None, constraints=None, maxtime=None, maxiters=None, method="asd"):
        # Get the name
        if name is None:
            name = "default"
        NamedItem.__init__(self, name)

        self.maxiters = maxiters  #: Maximum number of ASD iterations or hyperopt evaluations
        self.maxtime = maxtime  #: Maximum ASD time
        self.method = method  #: Optimization method name

        assert adjustments is not None, "Must specify some adjustments to carry out an optimization"
        assert measurables is not None, "Must specify some measurables to carry out an optimization"
        self.adjustments = sc.promotetolist(adjustments)
        self.measurables = sc.promotetolist(measurables)

        if constraints:
            self.constraints = [constraints] if not isinstance(constraints, list) else constraints
        else:
            self.constraints = None

        if adjustments is None or measurables is None:
            raise Exception("Must supply either a json or an adjustments+measurables")
        return

    def __repr__(self):
        return sc.prepr(self)

    def get_initialization(self, progset: ProgramSet, instructions: ProgramInstructions) -> tuple:
        """
        Get initial values for each adjustment

        The initial conditions depend nontrivially on both the progset and the instructions. Spending is
        present in the progset and optionally overwritten in the instructions. Therefore, it is necessary
        to check both when determining initial spending. Extraction of the initial values for each
        ``Adjustment`` is delegated to the ``Adjustment`` itself.

        Note also that the return arrays have length equal to the number of ``Adjustables``
        (since an ``Adjustment`` may contain several ``Adjustables``).

        :param progset: The program set to extract initial conditions from
        :param instructions: Instructions to extract initial conditions from
        :return: Tuple containing ``(initial,low,high)`` with arrays for
            - The initial value of each adjustable
            - The lower limit for each adjustable
            - The upper limit for each adjustable
        """

        # Return arrays of lower and upper bounds for each adjustable
        x0 = []
        for adjustment in self.adjustments:
            x0 += adjustment.get_initialization(progset, instructions)
        x0 = np.array(x0)
        xmin = np.zeros(x0.shape)
        xmax = np.zeros(x0.shape)

        ptr = 0
        for adjustment in self.adjustments:
            for adjustable in adjustment.adjustables:
                bounds = adjustable.get_hard_bounds(x0[ptr])
                xmin[ptr] = bounds[0]
                xmax[ptr] = bounds[1]
                if x0[ptr] > xmax[ptr]:
                    raise InvalidInitialConditions('Adjustment "%s" has an adjustable with initial value of %.2f but an upper bound of %.2f' % (adjustment.name, x0[ptr], xmax[ptr]))
                elif x0[ptr] < xmin[ptr]:
                    raise InvalidInitialConditions('Adjustment "%s" has an adjustable with initial value of %.2f but a lower bound of %.2f' % (adjustment.name, x0[ptr], xmax[ptr]))
                ptr += 1

        return x0, xmin, xmax

    def update_instructions(self, asd_values, instructions: ProgramInstructions) -> None:
        """
        Apply all Adjustments

        This method takes in a list of values (same length as number of adjustables) and iteratively
        calls each ``Adjustment`` in the optimization to update the instructions (in place)

        :param asd_values: A list of values
        :param instructions: The ``ProgramInstructions`` instance to update

        """

        idx = 0
        for adjustment in self.adjustments:
            adjustment.update_instructions(asd_values[idx : idx + len(adjustment.adjustables)], instructions)
            idx += len(adjustment.adjustables)

    def get_hard_constraints(self, x0, instructions: ProgramInstructions) -> list:
        """
        Get hard constraints

        This method calls ``get_hard_constraint`` on each ``Constraint`` in the ``Optimization``
        iteratively, and returns them as a list.

        Note that the initial optimization values ``x0`` are applied _before_ the hard constraint is computed.
        This ensures that the hard constraints are relative to the initial conditions in the optimization, not
        the initial instructions. For example, if a parametric overwrite is present, the hard constraint will
        be relative to whatever spending is produced by the initial values of the parametric overwrite.

        :param x0: The initial values for optimization - these are applied to the instructions prior to extracting hard constraints
        :param instructions: The initial instructions
        :return: A list of hard constraints, as many items as there are constraints

        """
        # Return hard constraints based on the starting initialization
        instructions = sc.dcp(instructions)
        self.update_instructions(x0, instructions)
        if not self.constraints:
            return list()
        else:
            return [x.get_hard_constraint(self, instructions) for x in self.constraints]

    def get_baselines(self, pickled_model) -> list:
        """
        Return Measurable baseline values

        This method is run at the start of the `optimize` script, and is used to
        retrieve the baseline values for the Measurable. Note that the baseline values are obtained
        based on the original instructions (stored in the pickled model), independent of the initial
        parameters used for optimization. The logic is that the initial parameters for the optimization
        are a choice dictated by the numerics of optimization (e.g. needing to start from a particular
        part of the parameter space) rather than anything intrinsic to the problem, whereas the
        initial instructions reflect the actual baseline conditions.

        :param pickled_model:
        :param x0: The initial parameter values
        :param hard_constraints: List of hard constraint values
        :return: A list of Measurable baseline values

        """

        model = pickle.loads(pickled_model)
        model.process()
        baselines = [m.get_baseline(model) for m in self.measurables]
        return baselines

    def get_adjustment(self, name: str) -> Adjustment:
        """
        Retrieve adjustment by name

        :param name:
        :return:
        """
        for adjustment in self.adjustments:
            if adjustment.name == name:
                return adjustment
        else:
            sc.suggest(name, [x.name for x in self.adjustments], die=True)
            raise NotFoundError(f'Adjustment "{name}" could not be found')

    def constrain_instructions(self, instructions: ProgramInstructions, hard_constraints: list) -> float:
        """
        Apply all constraints in-place, return penalty

        This method takes in the proposed instructions, and a list of hard constraints. Each constraint is
        applied to the instructions iteratively, passing in that constraint's own hard constraint, and the
        penalty is accumulated and returned.

        :param instructions: The current proposed ``ProgramInstructions``
        :param hard_constraints: A list of hard constraints the same length as the number of constraints
        :return: The total penalty value (if not finite, model integration will be skipped and the parameters will be rejected)

        """

        constraint_penalty = 0.0
        if self.constraints:
            for constraint, hard_constraint in zip(self.constraints, hard_constraints):
                constraint_penalty += constraint.constrain_instructions(instructions, hard_constraint, self)
        return constraint_penalty

    def compute_objective(self, model, baselines: list) -> float:
        """
        Return total objective function

        This method accumulates the objective values returned by each ``Measurable``, passing in
        the corresponding baseline values where required.

        :param model: A simulated ``Model`` object
        :param baselines: List of baseline values the same length as the number of ``Measurables``
        :return: The total/net objective value

        """

        # TODO - This method just adds the objective from each Measurable
        # It might be desirable in future to have more complex functions of the Measurable e.g.
        # sqrt of sum of squares. It's not clear yet whether this would be better as a transformation
        # applied here, or as a kind of meta-measurable. The former is probably simpler
        # to implement. But since there is no immediate need, this can be implemented later

        objective = 0.0
        for measurable, baseline in zip(self.measurables, baselines):
            objective += measurable.eval(model, baseline)
        return objective


def _objective_fcn(x, pickled_model, optimization, hard_constraints: list, baselines: list):
    """
    Return objective value

    This wrapper function takes in a vector of proposed parameters and returns the objective value.
    It is typically not called directly - instead, it is partialled to bind all arguments except ``x``
    and then passed to whichever optimization algorithm is used to optimize ``x``

    :param x: Vector of proposed parameter values
    :param pickled_model: A pickled ``Model`` - should contain a set of instructions
    :param optimization: An ``Optimization``
    :param hard_constraints: A list of hard constraints (should be the same length as ``optimization.constraints``)
    :param baselines: A list of measurable baselines (should be the same length as ``optimization.measurables``)
    :return:


    """

    try:
        model = pickle.loads(pickled_model)
        optimization.update_instructions(x, model.program_instructions)
        optimization.constrain_instructions(model.program_instructions, hard_constraints)
        model.process()
    except FailedConstraint:
        return np.inf  # Return an objective of `np.inf` if the constraints could not be satisfied by ``x``

    obj_val = optimization.compute_objective(model, baselines)

    # TODO - use constraint penalty somehow
    # The idea is to keep the optimization in a parameter regime where large corrections to the instructions
    # are not required. However, using ths constraint penalty directly can lead to it dominating the objective,
    # and with ASD's single-parameter stepping this prevents convergence. So needs some further work
    #
    # constaint_penalty = optimization.constrain_instructions(...)
    # obj_val += 0.0 * constraint_penalty

    return obj_val


def optimize(project, optimization, parset: ParameterSet, progset: ProgramSet, instructions: ProgramInstructions, x0=None, xmin=None, xmax=None, hard_constraints=None, baselines=None, optim_args: dict = None):
    """
    Main user entry point for optimization

    The optional inputs `x0`, `xmin`, `xmax` and `hard_constraints` are used when
    performing parallel optimization (implementation not complete yet), in which case
    they are computed by the parallel wrapper to `optimize()`. Normally these variables
    would not be specified by users, because they are computed from the `Optimization`
    together with the instructions (because relative constraints in the Optimization are
    interpreted as being relative to the allocation in the instructions).

    :param project: A :class:`Project` instance
    :param optimization: An :class:`Optimization` instance
    :param parset: A :class:`ParameterSet` instance
    :param progset: A :class:`ProgramSet` instance
    :param instructions: A :class:`ProgramInstructions` instance
    :param x0: Not for manual use - override initial values
    :param xmin: Not for manual use - override lower bounds
    :param xmax: Not for manual use - override upper bounds
    :param hard_constraints: Not for manual use - override hard constraints
    :param baselines: Not for manual use - override Measurable baseline values (for relative Measurables)
    :param optim_args: Pass a dictionary of keyword arguments to pass to the optimization algorithm (set in ``optimization.method``)
    :return: A :class:`ProgramInstructions` instance representing optimal instructions

    """

    assert optimization.method in ["asd", "pso", "hyperopt"]

    model = Model(project.settings, project.framework, parset, progset, instructions)
    pickled_model = pickle.dumps(model)  # Unpickling effectively makes a deep copy, so this _should_ be faster

    initialization = optimization.get_initialization(progset, model.program_instructions)
    x0 = x0 if x0 is not None else initialization[0]
    xmin = xmin if xmin is not None else initialization[1]
    xmax = xmax if xmax is not None else initialization[2]

    if not hard_constraints:
        hard_constraints = optimization.get_hard_constraints(x0, model.program_instructions)  # The optimization passed in here knows how to calculate the hard constraints based on the program instructions

    if not baselines:
        baselines = optimization.get_baselines(pickled_model)  # The optimization passed in here knows how to calculate the hard constraints based on the program instructions

    # Prepare additional arguments for the objective function
    args = {
        "pickled_model": pickled_model,
        "optimization": optimization,
        "hard_constraints": hard_constraints,
        "baselines": baselines,
    }

    # Check that the initial conditions are OK
    # Note that this cannot be done by `optimization.get_baselines` because the baselines need to be computed against the
    # initial instructions which might be different to the initial conditions (e.g. baseline spending vs the scaled-up
    # initialization used when minimizing spending)
    initial_objective = _objective_fcn(x0, **args)
    if not np.isfinite(initial_objective):
        raise InvalidInitialConditions("Optimization cannot begin because the objective function was %s for the specified initialization" % (initial_objective))

    if optimization.method == "asd":
        default_args = {
            "maxiters": optimization.maxiters,
            "maxtime": optimization.maxtime,
            "xmin": xmin,
            "xmax": xmax,
        }

        # Set ASD verbosity based on Atomica logging level
        log_level = logger.getEffectiveLevel()
        if log_level < logging.WARNING:
            default_args["verbose"] = 2
        else:
            default_args["verbose"] = 0

        optim_args = sc.mergedicts(default_args, optim_args)
        opt_result = sc.asd(_objective_fcn, x0, args, **optim_args)
        x_opt = opt_result["x"]

    elif optimization.method == "pso":

        import pyswarm

        default_args = {"maxiter": 3, "lb": xmin, "ub": xmax, "minstep": 1e-3, "debug": True}
        optim_args = sc.mergedicts(default_args, optim_args)

        if np.any(~np.isfinite(xmin)) or np.any(~np.isfinite(xmax)):
            errormsg = "PSO optimization requires finite upper and lower bounds to specify the search domain (i.e. every Adjustable needs to have finite bounds)"
            raise Exception(errormsg)

        x_opt, _ = pyswarm.pso(_objective_fcn, kwargs=args, **optim_args)

    elif optimization.method == "hyperopt":

        import hyperopt
        import functools

        if np.any(~np.isfinite(xmin)) or np.any(~np.isfinite(xmax)):
            errormsg = "hyperopt optimization requires finite upper and lower bounds to specify the search domain (i.e. every Adjustable needs to have finite bounds)"
            raise Exception(errormsg)

        space = []
        for i, (lower, upper) in enumerate(zip(xmin, xmax)):
            space.append(hyperopt.hp.uniform(str(i), lower, upper))
        fcn = functools.partial(_objective_fcn, **args)  # Partial out the extra arguments to the objective

        default_args = {"max_evals": optimization.maxiters if optimization.maxiters is not None else 100, "algo": hyperopt.tpe.suggest}
        optim_args = sc.mergedicts(default_args, optim_args)

        x_opt = hyperopt.fmin(fcn, space, **optim_args)
        x_opt = np.array([x_opt[str(n)] for n in range(len(x_opt.keys()))])

    elif callable(optimization.method):
        # Use custom optimization function
        # Placeholder functionality - not tested!

        optim_args = {} if optim_args is None else optim_args
        x_opt = optimization.method(_objective_fcn, **optim_args)

    else:
        raise Exception("Unrecognized optimization method")

    # Use the optimal parameter values to generate new instructions
    optimization.update_instructions(x_opt, model.program_instructions)
    optimization.constrain_instructions(model.program_instructions, hard_constraints)
    return model.program_instructions  # Return the modified instructions
    # Note that we do not return the value of the objective here because *in general* the objective isn't required
    # or expected to have a meaningful interpretation because it may arbitrarily combine quantities (e.g. spending
    # and epi outcomes) or is otherwise subject to the choice of weighting (e.g. impact vs equity). Therefore,
    # it is recommended that an _absolute_ metric be computed from the returned optimized instructions rather than
    # capturing the optimization objective. For example, in an optimization trading off equity against impact,
    # the deaths averted could be computed using the optimized instructions and returned as an absolute measure of quality.


# def parallel_optimize(project,optimization,parset,progset,program_instructions):
#
#    raise NotImplementedError
#
#    initialization = optimization.get_initialization(progset, program_instructions)
#    xmin = initialization[1]
#    xmax = initialization[2]
#
#    initial_instructions = program_instructions
#    for i in range(0,n_rounds):
#        for i in range(0,n_threads):
#            optimized_instructions[i] = optimize(optimization, parset, progset, initial_instructions, x0=None, xmin=xmin, xmax=xmax)
#        initial_instructions = pick_best(optimized_instructions)
#    return initial_instructions # Best one from the last round


def constrain_sum_bounded(x: np.array, s: float, lb: np.array, ub: np.array) -> np.array:
    """
    Bounded nearest constraint sum

    :param x: Array of proposed values to constrain
    :param s: Target value for ``sum(x)``
    :param lb: Array of lower bounds, same size as x
    :param ub: Array of upper bounds, same size as x
    :param tolerance: Absolute tolerance for the constrained sum ``s``
    :return: Array same size as ``x``, such that ``sum(x)==s`` and ``x[i]>=lb`` and ``x[i]<=ub`` for ``i<len(x)``
    :raises: :class:`FailedConstraint` if it was not possible to constrain
    """
    tolerance = 1e-6

    # Normalize values
    x0_scaled = x / (x.sum() or 1)  # Normalize the initial values, unless they sum to 0 (i.e., they are all zero)
    lb_scaled = lb / s
    ub_scaled = ub / s

    # First, check if the constraint is already satisfied just by multiplicative rescaling
    # The final check for x0_scaled.sum()==1 catches the case where all of the input values are 0
    if np.all((x0_scaled >= lb_scaled) & (x0_scaled <= ub_scaled)) and np.isclose(x0_scaled.sum(), 1):
        return x0_scaled * s

    # If not, we need to actually run the constrained optimization
    bounds = [(lower, upper) for lower, upper in zip(lb_scaled, ub_scaled)]

    def jacfcn(x):
        # Explicitly specify the Jacobian associated with changes in the allocation
        dist = np.linalg.norm(x - x0_scaled)
        if dist == 0:
            return np.zeros(x.shape)
        else:
            return (x - x0_scaled) / dist

    LinearConstraint = [{"type": "eq", "fun": lambda x: np.sum(x) - 1, "jac": lambda x: np.ones(x.shape)}]
    res = scipy.optimize.minimize(lambda x: np.linalg.norm(x - x0_scaled), x0_scaled, jac=jacfcn, bounds=bounds, constraints=LinearConstraint, method="SLSQP", options={"ftol": 1e-5, "maxiter": 1000})

    if not res["success"]:
        logger.warning("constrain_sum_bounded() failed - rejecting proposed parameters")
        raise FailedConstraint()

    # Enforce upper/lower bound constraints to prevent numerically exceeding them
    sol = np.minimum(np.maximum(res["x"], lb_scaled), ub_scaled) * s
    assert np.isclose(sol.sum(), s), f"FAILED as {sol} has a total of {sol.sum()} which is not sufficiently close to the target value {s}"
    return sol



import atomica as at
import numpy as np


def test_projectsettings():

    # Check that the sim end year is automatically updated if the
    # end year or the timestep is changed
    s = at.ProjectSettings(sim_start=2000, sim_end=2005, sim_dt=1)

    # Test changing sim end
    s.sim_end = 2006.5
    assert s.sim_end == 2007

    # Test changing dt
    s.sim_dt = 2
    assert s.sim_end == 2008

    # Test changing via update time vector
    s.update_time_vector(end=2009)
    assert s.sim_end == 2010

    # Test changing during construction
    s = at.ProjectSettings(sim_start=2000, sim_end=2001.1, sim_dt=0.25)
    assert s.sim_end == 2001.25

    # Check that the tvec returned is inclusive of the end year
    s = at.ProjectSettings(sim_start=2000, sim_end=2003, sim_dt=1)
    assert np.all(s.tvec == np.array([2000, 2001, 2002, 2003]))


if __name__ == "__main__":
    test_projectsettings()



import atomica as at
import numpy as np
import os

testdir = at.parent_dir()  # Must be relative to current file to work with tox


def test_program_coverage_calculation():

    # F = at.ProjectFramework(testdir+'test_program_calc_framework.xlsx')
    # D = at.ProjectData.new(F, tvec=[2019],pops=1, transfers=0)
    # D.save(testdir+'test_program_calc_databook.xlsx')
    P = at.Project(framework=testdir / "test_program_calc_framework.xlsx", databook=testdir / "test_program_calc_databook.xlsx", do_run=False)

    # ps = at.ProgramSet.new(project=P,progs=3,tvec=[2019])
    # ps.save(testdir+'test_program_calc_progbook.xlsx')
    P.load_progbook(testdir / "test_program_calc_progbook.xlsx")
    P.settings.update_time_vector(start=2019, end=2020, dt=0.25)

    ins = at.ProgramInstructions(start_year=2019)
    res = P.run_sim(parset=0, progset=0, progset_instructions=ins)

    # Check the resulting calculations
    assert res.get_variable("txrate1")[0].vals[0] == 0.8  # At this timestep, should have ANNUALIZED probability of 0.8
    assert res.get_variable("txrate2")[0].vals[0] == 80  # At this timestep, should have ANNUALIZED flow rate of 80

    assert res.get_variable("txrate1:flow")[0].vals[0] == 20  # At this timestep, should have timestep flow of 20
    assert res.get_variable("txrate2:flow")[0].vals[0] == 20  # At this timestep, should have timestep flow of 20

    # With 80 people eligible, funding to cover 20 people at this timestep, the coverage should be 25%
    # and we should continue to move 20 people. The annualized probability is now 1
    assert res.get_variable("txrate1")[0].vals[1] == 1  # At this timestep, should have ANNUALIZED probability of 1
    assert res.get_variable("txrate2")[0].vals[1] == 80  # At this timestep, should have ANNUALIZED flow rate of 80

    assert res.get_variable("txrate1:flow")[0].vals[1] == 20  # At this timestep, should have timestep flow of 20
    assert res.get_variable("txrate2:flow")[0].vals[1] == 20  # At this timestep, should have timestep flow of 20

    assert res.get_variable("prop_art")[0].vals[0] == 0.5  # Parameter value should correspond to 50% coverage

    # Test applying an overwrite
    ins2 = at.ProgramInstructions(start_year=2019, coverage={"tx_prob": 0.5, "tx_num": 0.5, "art": 0.4, "tx_cont": 0.0})
    res2 = P.run_sim(parset=0, progset=0, progset_instructions=ins2)

    assert res2.get_variable("txrate1")[0].vals[0] == 0.5  # At this timestep, should have ANNUALIZED probability of 0.5
    assert res2.get_variable("txrate2")[0].vals[0] == 50  # At this timestep, should have ANNUALIZED flow corresponding to 50% coverage i.e. 50 people

    assert res2.get_variable("txrate1:flow")[0].vals[0] == 12.5
    assert res2.get_variable("txrate2:flow")[0].vals[0] == 12.5

    # Actual flow rates should be the same regardless of whether the parameter was in number or probability units
    assert np.all(res2.get_variable("txrate1:flow")[0].vals == res2.get_variable("txrate2:flow")[0].vals)

    # Unlike with spending, the coverage remains fixed. Thus with 87.5 people eligible and a TIMESTEP coverage of 50/4=12.5% we have
    # a timestep number transferred of 10.9375. The parameter probability remains 0.5. For the number parameter, we have a timestep
    # coverage of 0.125 which gets multiplied by the parameter source popsize to get the parameter value. The source popsize is
    # now 87.5 so the parameter value is 87.5*0.125/0.25=43.75
    assert res2.get_variable("txrate1")[0].vals[1] == 0.5  # At this timestep, should have ANNUALIZED probability of 0.5
    assert res2.get_variable("txrate2")[0].vals[1] == 43.75  # At this timestep, should have ANNUALIZED flow corresponding to 50% coverage i.e. 50 people

    assert res2.get_variable("txrate1:flow")[0].vals[1] == 10.9375
    assert res2.get_variable("txrate2:flow")[0].vals[1] == 10.9375

    # Actual flow rates should be the same regardless of whether the parameter was in number or probability units
    assert np.all(res2.get_variable("txrate1:flow")[0].vals == res2.get_variable("txrate2:flow")[0].vals)

    assert res2.get_variable("prop_art")[0].vals[0] == 0.4  # Parameter value should correspond to 40% coverage

    # Coverage overwrite for continuous programs

    # Test if a continuous coverage program applies correctly - this should _not_ have a dt adjusted probability or number
    # Note that coverage that is emitted by `ProgramSet.get_prop_coverage` is
    ins3 = at.ProgramInstructions(start_year=2019, coverage={"tx_prob": 0.0, "tx_num": 0.0, "art": 0.4, "tx_cont": 1.0})
    res3 = P.run_sim(parset=0, progset=0, progset_instructions=ins3)

    assert res3.get_variable("txrate1")[0].vals[0] == 2.0  # At this timestep, should have ANNUALIZED probability of 2.0 (100% of coverage with an effect of 0.5)
    assert res3.get_variable("txrate2")[0].vals[0] == 200  # At this timestep, should have ANNUALIZED flow corresponding to 100% coverage at 0.5 effect i.e. 200 people/year

    assert res3.get_variable("txrate1:flow")[0].vals[0] == 50
    assert res3.get_variable("txrate2:flow")[0].vals[0] == 50

    # Actual flow rates should be the same regardless of whether the parameter was in number or probability units
    assert np.all(res3.get_variable("txrate1:flow")[0].vals == res3.get_variable("txrate2:flow")[0].vals)

    # Test the combination of continuous coverage and one-off coverage programs
    ins4 = at.ProgramInstructions(start_year=2019, coverage={"tx_prob": 0.5, "tx_num": 0.5, "art": 0.4, "tx_cont": 1.0})
    res4 = P.run_sim(parset=0, progset=0, progset_instructions=ins4)

    # 100% timestep coverage of the continuous program, with 0.5 effect
    # 50% annual coverage = 12.5% timestep coverage of tx_prob with 1.0 effect
    # Equates to 1*0.125+0.5*(1-0.125)=0.5625/timestep. Or 2.25 annually
    annual_rate = (1 * 0.125 + 0.5 * (1 - 0.125)) * 4
    assert res4.get_variable("txrate1")[0].vals[0] == annual_rate
    assert res4.get_variable("txrate2")[0].vals[0] == annual_rate * 100  # At this timestep, should have ANNUALIZED flow corresponding to the above

    assert res4.get_variable("txrate1:flow")[0].vals[0] == annual_rate * 100 / 4
    assert res4.get_variable("txrate2:flow")[0].vals[0] == annual_rate * 100 / 4

    # Actual flow rates should be the same regardless of whether the parameter was in number or probability units
    assert np.all(res4.get_variable("txrate1:flow")[0].vals == res4.get_variable("txrate2:flow")[0].vals)


#
# def test_no_program_progset():
#
#     P = at.demo('sir', do_run=False)
#
#     tvec = P.settings.tvec
#     dt = P.settings.sim_dt
#     instructions = at.ProgramInstructions(start_year=2020)
#
#
#     ps1 = P.progsets[0]
#     ps2 = at.ProgramSet(framework=P.framework, data=P.data)
#
#     capacities1 = ps1.get_capacities(tvec, dt, instructions)
#     capacities2 = ps2.get_capacities(tvec, dt, instructions)
#
#     def get_prop_coverage(self, tvec, dt, capacities: dict, num_eligible: dict, instructions=None) -> dict:
#
#
#     coverage1 = ps1.get_coverage(tvec, dt, capacities1, )
#     ps.to_spreadsheet() # Check that a spreadsheet can be produced
#     ps
#
#
#     ps1 = P.progsets[0]
#   _get_code_name()    add_pop()           remove_par()
#   _normalize_inpu...  add_program()       remove_pop()
#   _read_effects()     copy()              remove_program()
#   _read_spending()    from_spreadsheet()  sample()
#   _read_targeting()   get_alloc()         save()
#   _write_effects()    get_capacities()    to_spreadsheet()
#   _write_spending()   get_outcomes()      to_workbook()
#   _write_targeting()  get_prop_covera...  validate()
#   add_comp()          new()               add_par()
#   remove_comp()
#
# TypeError: get_prop_coverage() missing 4 required positional arguments: 'tvec', 'dt', 'capacities', and 'num_eligible'


if __name__ == "__main__":
    test_program_coverage_calculation()
    # test_no_program_progset()



# This script checks that the junction update is correct

import numpy as np
import atomica as at
import matplotlib.pyplot as plt
import os


def test_seasonal():
    testdir = at.parent_dir()
    tmpdir = testdir / "temp"

    F_path = testdir / "framework_seasonal_test.xlsx"
    D_path = tmpdir / "databook_seasonal_test.xlsx"

    F = at.ProjectFramework(F_path)
    D = at.ProjectData.new(F, np.arange(2000, 2001), pops={"pop1": "Population 1"}, transfers=0)
    D.save(D_path)

    P = at.Project(name="test", framework=F, do_run=False)
    P.settings.update_time_vector(end=2005, dt=0.02)
    P.load_databook(databook_path=D_path, make_default_parset=True, do_run=True)
    r = P.results[0]

    d = at.PlotData(r, outputs=["seasonal_max"], project=P)
    at.plot_series(d)
    plt.title("Seasonal rainfall")

    d = at.PlotData(r, outputs=["seasonal_jan", "seasonal_jun"], project=P)
    at.plot_series(d)
    plt.title("Seasonal rainfall")

    d = at.PlotData(r, outputs=["birth"], project=P)
    at.plot_series(d)

    d = at.PlotData(r, outputs=["mos"], project=P)
    at.plot_series(d)


if __name__ == "__main__":
    test_seasonal()
